import os
import threading
import logging
from functools import partial

from six.moves.queue import Queue, Empty
import six
from threading import Thread

import koji
from more_executors import Executors
from more_executors.futures import f_map

from ..source import Source
from ..model import (
    KojiBuildInfo,
    RpmPushItem,
    ModuleMdPushItem,
    ModuleMdSourcePushItem,
    OperatorManifestPushItem,
    ContainerImagePushItem,
    SourceContainerImagePushItem,
)
from ..helpers import list_argument, try_int, as_completed_with_timeout_reset
from .modulemd import Module
from .koji_containers import ContainerArchiveHelper, MIME_TYPE_MANIFEST_LIST

LOG = logging.getLogger("pushsource")
CACHE_LOCK = threading.RLock()

# Arguments used for retry policy.
# Provided so it can be overridden from tests to reduce time spent on retries.
RETRY_ARGS = {}


class ListArchivesCommand(object):
    def __init__(self, build):
        self.build = build
        self.call = None

    def execute(self, source, session):
        archives_cache = source._cache.setdefault("archives", {})
        ident = self.build["id"]
        if ident in archives_cache:
            return 0

        LOG.debug("Get koji archives %s", ident)
        self.call = session.listArchives(ident)
        return 1

    def save(self, source, _):
        if self.call is not None:
            source._cache["archives"][self.build["id"]] = self.call.result
            source._cache["archives"][self.build["nvr"]] = self.call.result


class GetBuildCommand(object):
    def __init__(self, ident, list_archives=False):
        self.ident = ident
        self.list_archives = list_archives
        self.call = None

    def execute(self, source, session):
        build_cache = source._cache.setdefault("build", {})
        if self.ident in build_cache:
            return 0
        LOG.debug("Get koji build %s", self.ident)
        self.call = session.getBuild(self.ident)
        return 1

    def save(self, source, koji_queue):
        if self.call is None:
            build = source._cache["build"][self.ident]
        else:
            # Build is saved under both NVR and ID
            build = self.call.result
            if build:
                source._cache["build"][build["id"]] = build
                source._cache["build"][build["nvr"]] = build
            else:
                source._cache["build"][self.ident] = build
        if self.list_archives and build:
            koji_queue.put(ListArchivesCommand(build))


class GetRpmCommand(object):
    def __init__(self, ident):
        self.ident = ident
        self.call = None

    def execute(self, source, session):
        rpm_cache = source._cache.setdefault("rpm", {})
        if self.ident in rpm_cache:
            return 0
        LOG.debug("Get koji RPM %s", self.ident)
        self.call = session.getRPM(self.ident)
        return 1

    def save(self, source, koji_queue):
        if self.call is not None:
            source._cache["rpm"][self.ident] = self.call.result
        # We have to get the RPM's build as well.
        if source._cache["rpm"][self.ident]:
            build_id = source._cache["rpm"][self.ident]["build_id"]
            koji_queue.put(GetBuildCommand(build_id))


class KojiSource(Source):
    """Uses koji artifacts as the source of push items."""

    _BATCH_SIZE = int(os.environ.get("PUSHSOURCE_KOJI_BATCH_SIZE", "100"))

    def __init__(
        self,
        url,
        dest=None,
        rpm=None,
        module_build=None,
        module_filter_filename=None,
        container_build=None,
        signing_key=None,
        basedir=None,
        threads=4,
        timeout=60 * 30,
        cache=None,
        executor=None,
    ):
        """Create a new source.

        Parameters:
            url (str)
                URL of the XML-RPC endpoint for koji hub.

            dest (str, list[str])
                The destination(s) to fill in for push items created by this
                source. If omitted, all push items have empty destinations.

            rpm (list[str, int])
                RPM identifier(s). Can be koji IDs (integers) or filenames.
                The source will yield all RPMs identified by this list.

            module_build (list[str, int])
                Build identifier(s). Can be koji IDs (integers) or build NVRs.
                The source will yield modulemd files belonging to these
                build(s), filtered by ``module_filter_filename`` (if given).

            module_filter_filename (list[str])
                Filename(s) of modulemd archives to include in output.

                When ``module_build`` is given, modulemd files are located as
                koji archives with names of the form ``modulemd.<arch>.txt``.
                By default, all of these archives will be processed.

                Providing a non-empty list for ``module_filter_filename`` will
                instead only process modulemd files with those filenames. This
                can be used to select only certain arches.

                Has no effect if ``module_build`` is not provided.

            container_build (list[str, int])
                Build identifier(s). Can be koji IDs (integers) or build NVRs.
                The source will yield all container images produced by these
                build(s), along with associated artifacts such as operator manifest
                bundles.

            signing_key (list[str])
                GPG signing key ID(s). If provided, content must be signed
                using one of the provided keys. Include ``None`` if unsigned
                should also be permitted.

                Keys should be listed in the order of preference.

            basedir (str)
                Base directory of koji content (e.g. /mnt/koji).
                This directory must be readable locally.

            threads (int)
                Number of threads used for concurrent queries to koji.

            timeout (int)
                Number of seconds after which an error is raised, if no progress is
                made during queries to koji.

            cache (dict)
                A cache used to retain the results of XML-RPC calls, to avoid
                repeated calls.

                Providing a cache only gives a benefit if the current process
                will create more than a single instance of KojiSource. In this
                case, some calls may be avoided by passing the same cache
                to each instance.

            executor (concurrent.futures.Executor)
                A custom executor used to submit calls to koji.
        """
        self._url = url
        self._rpm = [try_int(x) for x in list_argument(rpm)]
        self._module_build = [try_int(x) for x in list_argument(module_build)]
        self._module_filter_filename = list_argument(
            module_filter_filename, retain_none=True
        )
        self._container_build = [try_int(x) for x in list_argument(container_build)]
        self._signing_key = list_argument(signing_key)
        self._dest = list_argument(dest)
        self._timeout = timeout
        self._pathinfo = koji.PathInfo(basedir)
        self._cache = {} if cache is None else cache
        self._threads = threads
        self._executor = (
            executor
            or Executors.thread_pool(name="pushsource-koji", max_workers=threads)
            .with_retry(**RETRY_ARGS)
            .with_cancel_on_shutdown()
        )

        self._on_shutdown = []
        if not executor:
            self._on_shutdown.append(lambda: self._executor.shutdown(True))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for cb in self._on_shutdown:
            cb()

    @property
    def _koji_session(self):
        # A koji client session.
        # Each thread uses a separate client.
        with CACHE_LOCK:
            tls = self._cache.setdefault("tls", threading.local())
            if not hasattr(tls, "koji_session"):
                LOG.debug("Creating koji session: %s", self._url)
                tls.koji_session = koji.ClientSession(self._url)
        return tls.koji_session

    def _koji_check(self):
        # Do a basic connection check with koji.
        # If this succeeds, we can be reasonably sure that the koji connection is
        # more or less working.
        try:
            version = self._executor.submit(self._koji_get_version).result()
        except Exception as ex:  # pylint: disable=broad-except
            # TODO: drop this log when py2 support is dropped.
            # It's only here because py2 has no exception chaining.
            msg = "Communication error with koji at %s" % self._url
            LOG.exception(msg)

            six.raise_from(RuntimeError(msg), ex)

        LOG.debug("Connected to koji %s at %s", version, self._url)

    def _koji_get_version(self):
        with CACHE_LOCK:
            if "koji_version" not in self._cache:
                self._cache["koji_version"] = self._koji_session.getKojiVersion()
            return self._cache["koji_version"]

    def _get_rpm(self, rpm):
        return self._cache["rpm"][rpm]

    def _get_build(self, build_id):
        return self._cache["build"][build_id]

    def _get_archives(self, build_id):
        return self._cache["archives"][build_id]

    def _push_items_from_rpm_meta(self, rpm, meta):
        LOG.debug("RPM metadata for %s: %s", rpm, meta)

        notfound = [RpmPushItem(name=str(rpm), dest=self._dest, state="NOTFOUND")]

        if not meta:
            LOG.error("RPM not found in koji: %s", rpm)
            return notfound

        build = self._get_build(meta["build_id"])
        build_path = self._pathinfo.build(build)
        rpm_path = None
        unsigned_path = os.path.join(build_path, self._pathinfo.rpm(meta))
        rpm_signing_key = None

        candidate_paths = []

        # If signing keys requested, try them in order of preference
        for key in self._signing_key:
            if key:
                key = key.lower()
                candidate = os.path.join(build_path, self._pathinfo.signed(meta, key))
            else:
                candidate = unsigned_path
            candidate_paths.append(candidate)
            if os.path.exists(candidate):
                rpm_path = candidate
                rpm_signing_key = key
                break

        if self._signing_key:
            # If signing keys requested: we either found an RPM above, or an error occurs
            if not rpm_path:
                LOG.error(
                    "RPM not found in koji at path(s): %s", ", ".join(candidate_paths)
                )
                return notfound
        else:
            # If no signing key requested, we just use the calculated path directly
            # and don't even test for its existence. Note: absence of test for existence
            # is ported from old code and is not necessarily the best way.
            rpm_path = unsigned_path

        return [
            RpmPushItem(
                name=os.path.basename(rpm_path),
                src=rpm_path,
                dest=self._dest,
                signing_key=rpm_signing_key,
                build=build["nvr"],
            )
        ]

    def _module_filtered(self, file_path):
        ok_names = self._module_filter_filename
        filename = os.path.basename(file_path)
        if ok_names is not None and filename not in ok_names:
            LOG.debug("Skipping module %s due to module_filter_filename", file_path)
            return True

        return False

    def _get_module_name(self, nvr, file_path):
        # Return the best value for "name" we can calculate for a given modulemd
        # push item. In practice this means setting the push item name to NSVCA,
        # if the modulemd file exists and can be parsed as a single module.
        # NSVCA is a more meaningful name for the item than modulemd.<arch>.txt,
        # which will be the same for all modules of the same arch.
        #
        # Why only "if the modulemd file exists"? Well, it's only going to exist
        # if the caller has mounted the NFS volume associated with koji. This will
        # be true in production, and we could give a fatal error when this is not
        # the case, but it seems like a very harsh requirement. e.g. most of the
        # docs for this library are using fedora koji as an example, but nobody
        # running outside of fedora infra will be able to mount that volume,
        # meaning attempting to use the examples will immediately hit a brick wall.
        #
        # So, this is a compromise. If the file exists, it has to be valid, and we
        # use it; this gives the best behavior in production. If the file doesn't
        # exist, that's OK, and that's a compromise to avoid making the dev and test
        # requirements too onerous.
        basename = os.path.basename(file_path)

        if basename == "modulemd.src.txt":
            # These are modulemd packager files and not the built files with NSVCA,
            # do not attempt to parse
            return basename

        if not os.path.exists(file_path):
            # Don't have the file, don't attempt to parse
            return basename

        try:
            return Module.from_file(file_path).nsvca
        except:
            # If we fail, make it clear what we were attempting to do before raising
            LOG.exception(
                "In koji build %s, cannot load module metadata from %s", nvr, file_path
            )
            raise

    def _push_items_from_module_build(self, nvr, meta):
        LOG.debug("Looking for modulemd on %s, %s", nvr, meta)

        if not meta:
            message = "Module build not found in koji: %s" % nvr
            LOG.debug(message)
            raise ValueError(message)

        build_id = meta["id"]
        archives = self._get_archives(build_id)
        modules = [elem for elem in archives if elem.get("btype") == "module"]

        out = []
        for module in modules:
            file_path = os.path.join(
                self._pathinfo.typedir(meta, "module"), module["filename"]
            )

            if self._module_filtered(file_path):
                continue

            name = self._get_module_name(nvr, file_path)

            klass = ModuleMdPushItem
            if os.path.basename(file_path) == "modulemd.src.txt":
                klass = ModuleMdSourcePushItem

            out.append(
                klass(name=name, src=file_path, dest=self._dest, build=meta["nvr"])
            )
        return out

    def _push_items_from_container_build(self, nvr, meta):
        LOG.debug("Looking for container images on %s, %s", nvr, meta)

        if not meta:
            message = "Container image build not found in koji: %s" % nvr
            LOG.debug(message)
            raise ValueError(message)

        # The metadata we are interested in is documented here:
        # https://github.com/containerbuildsystem/atomic-reactor/blob/061d92e63cf27ae030e8ceed388ec34f51afb17b/docs/koji.md#type-specific-metadata
        extra = meta.get("extra") or {}
        typeinfo = extra.get("typeinfo") or {}
        # Per above doc, it is preferred to use the metadata under 'typeinfo' if present
        image = typeinfo.get("image") or extra.get("image")

        if image is None:
            message = "Build %s not recognized as a container image build" % nvr
            LOG.debug(message)
            raise ValueError(message)

        build_id = meta["id"]

        archives = self._get_archives(build_id)

        image_archives = [
            elem
            for elem in archives
            if elem.get("btype") == "image" and elem.get("type_name") == "tar"
        ]

        media_types = image.get("media_types") or []
        if MIME_TYPE_MANIFEST_LIST in media_types:
            # in manifest list case it is OK to have multiple archives (different arches).
            pass
        elif len(image_archives) != 1:
            # If we don't have a manifest list, we expect to have an image for one arch.
            # If there's not exactly one, it's unclear what's going on and we'd better halt.
            message = (
                "Could not find (exactly) one container image archive on koji build %s"
                % nvr
            )
            LOG.debug(message)
            raise ValueError(message)

        out = []

        for archive in image_archives:
            path = self._pathinfo.typedir(meta, archive["btype"])
            item_src = os.path.join(path, archive["filename"])

            helper = ContainerArchiveHelper(meta, archive)

            klass = ContainerImagePushItem
            if image.get("sources_for_nvr"):
                # This is a source image.
                klass = SourceContainerImagePushItem

            out.append(
                klass(
                    # TODO: name might be changed once we start parsing
                    # the metadata from atomic-reactor.
                    name=archive["filename"],
                    dest=self._dest,
                    src=item_src,
                    build=nvr,
                    # Note, we should be able to use the default KojiBuild
                    # construction from NVR here. The reason we don't is that
                    # there is some "impossible" test data in Pub which has
                    # brew builds whose NVR does not match their (n, v, r)
                    # in the build data and it's quite tough to fix that now.
                    build_info=KojiBuildInfo(
                        name=meta["name"],
                        version=meta["version"],
                        release=meta["release"],
                    ),
                    source_tags=helper.source_tags,
                    labels=helper.labels,
                    arch=helper.arch,
                    pull_info=helper.pull_info,
                )
            )

        # If this build had any operator-manifests archive, add that too.
        operator = self._get_operator_item(nvr, meta, archives)
        if operator:
            out.append(operator)

        return out

    def _get_operator_item(self, nvr, meta, archives):
        extra = meta.get("extra") or {}
        typeinfo = extra.get("typeinfo") or {}
        operator_manifests = typeinfo.get("operator-manifests") or {}
        archive_name = operator_manifests.get("archive")

        if not archive_name:
            # Try legacy form
            archive_name = extra.get("operator_manifests_archive")

        if not archive_name:
            # No operator manifests on this build
            return

        operator_archive = [a for a in archives if a["filename"] == archive_name]
        if len(operator_archive) != 1:
            message = (
                "koji build %s metadata refers to missing operator-manifests "
                'archive "%s"'
            ) % (nvr, archive_name)
            raise ValueError(message)

        archive = operator_archive[0]
        path = self._pathinfo.typedir(meta, archive["btype"])
        item_src = os.path.join(path, archive["filename"])

        return OperatorManifestPushItem(
            name=os.path.join(nvr, archive_name),
            dest=self._dest,
            src=item_src,
            build=nvr,
        )

    def _rpm_futures(self):
        # Get info from each requested RPM.
        rpm_fs = [(self._executor.submit(self._get_rpm, rpm), rpm) for rpm in self._rpm]

        # Convert them to lists of push items
        return [
            f_map(f, partial(self._push_items_from_rpm_meta, rpm)) for f, rpm in rpm_fs
        ]

    def _modulemd_futures(self):
        # Get info from each requested module build.
        module_build_fs = [
            (self._executor.submit(self._get_build, build), build)
            for build in self._module_build
        ]

        # Convert them to lists of push items
        return [
            f_map(f, partial(self._push_items_from_module_build, build))
            for f, build in module_build_fs
        ]

    def _container_futures(self):
        # Get info from each requested container build.
        build_fs = [
            (self._executor.submit(self._get_build, build), build)
            for build in self._container_build
        ]

        # Convert them to lists of push items
        return [
            f_map(f, partial(self._push_items_from_container_build, build))
            for f, build in build_fs
        ]

    def _do_fetch(self, koji_queue, exceptions):
        try:
            done = False

            while not done:
                count = 0
                pending_commands = []

                session = self._koji_session.multicall(
                    strict=True, batch=self._BATCH_SIZE
                )

                while not done:
                    try:
                        command = koji_queue.get_nowait()
                        count += command.execute(self, session)
                        pending_commands.append(command)
                    except Empty:
                        done = True
                        LOG.debug("koji fetch queue emptied")

                LOG.debug("koji multicall: about to execute %s call(s)", count)

                session.call_all()

                LOG.debug("koji multicall: executed %s call(s)", count)

                # multicall is now done.
                # Save the result to cache from each command.
                # This could potentially result in new queue entries.
                for command in pending_commands:
                    command.save(self, koji_queue)

                # If there were any commands processed, queue might no longer be
                # empty, so re-check it
                if pending_commands:
                    done = False
        except Exception as e:
            LOG.exception("Error during koji fetches")
            exceptions.append(e)
            raise

    def __iter__(self):
        # Try a (blocking) call to koji before anything else.
        #
        # The reasons for this are:
        #
        # - give an error which is hopefully easier to understand if the
        #   koji connection is completely broken (e.g. caller provides bogus
        #   URL)
        #
        # - work around pyasn1 bug https://github.com/etingof/pyasn1/issues/53
        #   which affects some (ancient) environments. That thread-safety bug
        #   can cause parsing of subjectAltName to fail if it first occurs from
        #   multiple threads simultaneously. Doing a warm-up from one thread
        #   before we spawn multiple threads is a way to avoid this.
        #
        self._koji_check()

        # Queue holding all requests we need to make to koji.
        # We try to fetch as much as we can early to make efficient use
        # of multicall.
        koji_queue = Queue()

        # We'll need to obtain all RPMs referenced by filename
        for rpm_filename in self._rpm:
            koji_queue.put(GetRpmCommand(ident=rpm_filename))

        # We'll need to obtain all builds from which we want modules,
        # as well as the archives from those
        for build_id in self._module_build:
            koji_queue.put(GetBuildCommand(ident=build_id, list_archives=True))

        # We'll need to obtain all container image builds
        for build_id in self._container_build:
            koji_queue.put(GetBuildCommand(ident=build_id, list_archives=True))

        # Put some threads to work on the queue.
        fetch_exceptions = []
        fetch_threads = [
            Thread(
                name="koji-%s-fetch-%s" % (id(self), i),
                target=self._do_fetch,
                args=(koji_queue, fetch_exceptions),
            )
            for i in range(0, self._threads)
        ]

        # Wait for all fetches to finish
        for t in fetch_threads:
            t.start()

        # If we are asked to shut down early then also shut down those threads
        # as soon as we can.
        def fetch_shutdown():
            try:
                # Empty the queue
                while True:
                    koji_queue.get_nowait()
            except Empty:
                pass

            # Then ask the threads to join
            for t in fetch_threads:
                t.join()

        self._on_shutdown.append(fetch_shutdown)

        for t in fetch_threads:
            t.join(self._timeout)

        # Re-raise exceptions, if any.
        # If we got more than one, we're only propagating the first.
        if fetch_exceptions:
            raise fetch_exceptions[0]

        # The queue must be empty now
        assert koji_queue.empty()

        push_items_fs = (
            self._modulemd_futures() + self._rpm_futures() + self._container_futures()
        )

        completed_fs = as_completed_with_timeout_reset(
            push_items_fs, timeout=self._timeout
        )
        for f in completed_fs:
            # If an exception occurred, this is where it will be raised.
            for pushitem in f.result():
                yield pushitem


Source._register_backend_builtin("koji", KojiSource)



from pushsource._impl import Source, SourceUrlError
from pushsource._impl.model import (
    PushItem,
    KojiBuildInfo,
    FilePushItem,
    CompsXmlPushItem,
    ModuleMdPushItem,
    ModuleMdSourcePushItem,
    ProductIdPushItem,
    RpmPushItem,
    ContainerImagePushItem,
    ContainerImagePullSpec,
    ContainerImageDigestPullSpec,
    ContainerImageTagPullSpec,
    ContainerImagePullInfo,
    SourceContainerImagePushItem,
    OperatorManifestPushItem,
    AmiPushItem,
    AmiRelease,
    AmiBillingCodes,
    ErratumPushItem,
    ErratumReference,
    ErratumModule,
    ErratumPackage,
    ErratumPackageCollection,
)

from pushsource._impl.backend import (
    ErrataSource,
    KojiSource,
    StagedSource,
    RegistrySource,
)



import hashlib
import os
import logging

from frozenlist2 import frozenlist

from .. import compat_attr as attr
from .conv import (
    md5str,
    sha256str,
    upper_if_str,
    instance_of_str,
    instance_of,
    optional_str,
)


LOG = logging.getLogger("pushsource")
CHUNKSIZE = int(os.environ.get("PUSHSOURCE_CHUNKSIZE") or 1024 * 1024 * 16)


@attr.s()
class KojiBuildInfo(object):
    """A representation of a koji build."""

    name = attr.ib(type=str, validator=instance_of_str)
    """'name' component of build's NVR.

    Example: in "kf5-kio-5.83.0-2.el8.next", the name is "kf5-kio".
    """

    version = attr.ib(type=str, validator=instance_of_str)
    """'version' component of build's NVR.

    Example: in "kf5-kio-5.83.0-2.el8.next", the version is "5.83.0".
    """

    release = attr.ib(type=str, validator=instance_of_str)
    """'release' component of build's NVR.

    Example: in "kf5-kio-5.83.0-2.el8.next", the release is "2.el8.next".
    """

    @classmethod
    def _from_nvr(cls, nvr_str):
        if not nvr_str:
            return

        # Initial state: 'kf5-kio-5.83.0-2.el8.next'
        # Reverse:       'txen.8le.2-0.38.5-oik-5fk'
        nvr_rev = nvr_str[::-1]

        # Split:         ['txen.8le.2', '0.38.5', 'oik-5fk']
        rvn_rev = nvr_rev.split("-", 2)

        # Unreverse:     ['2.el8.next', '5.83.0', 'kf5-kio']
        (r, v, n) = [s[::-1] for s in rvn_rev]

        return cls(name=n, version=v, release=r)


@attr.s()
class PushItem(object):
    """A PushItem represents a single piece of content to be published.
    This may be an RPM, an advisory, a generic file, and so on.
    """

    name = attr.ib(type=str, validator=instance_of_str)
    """A name for this push item.

    A push item's name may be any short identifying string which is meaningful
    to end-users.

    In the common case of a regular file, a push item name will simply be a filename,
    optionally including leading path components.

    In all cases, a non-empty name must be provided.
    """

    state = attr.ib(type=str, default="PENDING")
    """The state of this push item.

    In the majority of cases, push sources will produce push items with
    a state of ``PENDING``, with the caller expected to later evolve the state.

    In rare cases, push sources may produce push items with other states, such
    as ``INVALIDFILE`` to indicate that the source located invalid content.
    """

    src = attr.ib(type=str, default=None, validator=optional_str)
    """The source of this push item.

    If the push item is a file, this will be the full path to the file.
    If the push item does not represent a file, this will generally be omitted.
    """

    dest = attr.ib(type=list, default=attr.Factory(frozenlist), converter=frozenlist)
    """Destination of this push item.

    The meaning of "dest" differs depending on the source used and its configuration.

    Examples for possible uses of "dest" include:

    * a path to a directory (for items pushed using rsync)
    * a Pulp repository name (for items pushed using Pulp)

    :type: list[str]
    """

    md5sum = attr.ib(type=str, default=None, converter=md5str)
    """Hex digest of MD5 checksum for this push item, if available.

    .. seealso::

        :meth:`with_checksums`
    """

    sha256sum = attr.ib(type=str, default=None, converter=sha256str)
    """Hex digest of SHA256 checksum for this push item, if available.

    .. seealso::

        :meth:`with_checksums`
    """

    origin = attr.ib(type=str, default=None, validator=optional_str)
    """A string representing the origin of this push item.

    The "origin" field is expected to record some info on how this push item
    was discovered. The exact semantics depend on the specific source backend
    in use.
    """

    build = attr.ib(type=str, default=None, validator=optional_str)
    """NVR for the koji build from which this push item was extracted, if any.

    .. seealso::

        :meth:`build_info`, which provides the same info in parsed form.
    """

    build_info = attr.ib(
        type=KojiBuildInfo, validator=instance_of((KojiBuildInfo, type(None)))
    )
    """Basic info on the koji build from which this push item was extracted, if any."""

    @build_info.default
    def _default_build_info(self):
        return KojiBuildInfo._from_nvr(self.build)

    signing_key = attr.ib(
        type=str, default=None, validator=optional_str, converter=upper_if_str
    )
    """If this push item was GPG signed, this should be an identifier for the
    signing key used.

    Generally a short key ID such as "F21541EB" is used, though the library
    doesn't enforce this.
    """

    def with_checksums(self):
        """Return a copy of this push item with checksums present.

        Many pushsource backends will produce push items with empty checksum attributes,
        since checksum calculation may be expensive and may not be required in all cases.

        The rule of thumb is that a pushsource will only provide checksums if they were
        available in metadata, so that reading the push item's file was unnecessary.
        For example:

        - with ``errata`` push source, checksums are generally available by default, because
          Errata Tool itself provides checksums on the files contained in an advisory.
        - with ``staged`` push source, checksums are unavailable by default, because the
          sums can only be calculated by reading file content from the staging area.

        Where checksums are needed, this utility method may be used to ensure they are present
        regardless of the push source used.

        This method entails reading the entire content of the file referenced
        by this push item, and may be:

        - slow: may need to read a large file from a remote system, e.g. from an NFS volume.
        - error-prone: reads from a remote system might fail.

        As such, when dealing with a large number of push items, you may want to consider
        using multiple threads to parallelize calls to ``with_checksums``, and retrying
        failing operations.

        If checksums are already present or if this item does not reference a file, this
        method is a no-op and returns the current push item, unmodified.

        Returns:
            :class:`~pushsource.PushItem`
                A copy of this item, guaranteed either to have non-empty :meth:`md5sum` and
                :meth:`sha256sum` attributes, or an empty :meth:`src` attribute (denoting that
                the item does not reference a file).

        .. versionadded:: 1.2.0
        """
        if not self.src:
            return self

        hashers = []

        if not self.md5sum:
            hashers.append((hashlib.new("md5"), "md5sum"))
        if not self.sha256sum:
            hashers.append((hashlib.new("sha256"), "sha256sum"))

        if not hashers:
            # Nothing to do
            return self

        LOG.debug("Start read: %s", self.src)

        with open(self.src, "rb") as src_file:
            while True:
                chunk = src_file.read(CHUNKSIZE)
                if not chunk:
                    break
                for (hasher, _) in hashers:
                    hasher.update(chunk)

        LOG.debug("End read: %s", self.src)

        updated_sums = {}
        for (hasher, attribute) in hashers:
            updated_sums[attribute] = hasher.hexdigest()

        return attr.evolve(self, **updated_sums)



from ...model import CompsXmlPushItem
from .staged_base import StagedBaseMixin, handles_type


class StagedCompsXmlMixin(StagedBaseMixin):
    @handles_type("COMPS")
    def __push_item(self, leafdir, _, entry):
        return CompsXmlPushItem(
            name=entry.name, src=entry.path, origin=leafdir.topdir, dest=[leafdir.dest]
        )



import datetime

from frozenlist2 import frozenlist

from .base import PushItem
from .. import compat_attr as attr
from .conv import datestr, instance_of_str, instance_of, optional_str, optional


@attr.s()
class AmiRelease(object):
    """Release metadata associated with an AMI."""

    product = attr.ib(type=str, validator=instance_of_str)
    """A short product name, for example "RHEL" if this is an image for Red Hat
    Enterprise Linux."""

    date = attr.ib(type=datetime.date, converter=datestr)
    """Date at which this image was generated."""

    arch = attr.ib(type=str, validator=instance_of_str)
    """Architecture of the image, for example "x86_64"."""

    respin = attr.ib(type=int, validator=instance_of(int))
    """Respin count. 0 for original build of an image, incremented by one for
    each rebuild."""

    version = attr.ib(type=str, default=None, validator=optional_str)
    """A <major>.<minor> version string for the image's product version, for example
    "7.9" if this is an image for Red Hat Enterprise Linux 7.9.x."""

    base_product = attr.ib(type=str, default=None, validator=optional_str)
    """For layered products, name of the base product for which the image should be used."""

    base_version = attr.ib(type=str, default=None, validator=optional_str)
    """For layered products, version of the base product for which the image should be used."""

    variant = attr.ib(type=str, default=None, validator=optional_str)
    """Variant of this image's product (only for products which have variants).  For example,
    "Server", for Red Hat Enterprise Linux Server."""

    type = attr.ib(type=str, default=None, validator=optional_str)
    """Release type, typically "ga" or "beta"."""


@attr.s()
class AmiBillingCodes(object):
    """Billing codes associated with an AMI."""

    name = attr.ib(type=str, default=None, validator=instance_of_str)
    """Billing codes name, for example Hourly2, arbitrary string for making image name unique."""

    codes = attr.ib(type=list, default=attr.Factory(frozenlist), converter=frozenlist)
    """List of billing codes, for example ['bp-1234abcd', 'bp-5678efgh'].

    :type: list[str]
    """


@attr.s()
class AmiPushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing an Amazon Machine Image (or "AMI").

    For push items of this type, the :meth:`~pushsource.PushItem.src` attribute
    refers to a file containing an EBS snapshot.  The push item contains additional
    metadata which can be used to associate the snapshot with an image.

    This library does not verify that the referenced file is a valid
    snapshot.
    """

    release = attr.ib(
        type=AmiRelease, default=None, validator=optional(instance_of(AmiRelease))
    )
    """Release metadata associated with this image."""

    type = attr.ib(type=str, default=None, validator=optional_str)
    """Billing type associated with the image, e.g. "hourly" or "access"."""

    region = attr.ib(type=str, default=None, validator=instance_of_str)
    """Region to which this image should be pushed."""

    virtualization = attr.ib(type=str, default=None, validator=instance_of_str)
    """Virtualization type, e.g. "hvm" or "pv"."""

    volume = attr.ib(type=str, default=None, validator=instance_of_str)
    """EBS volume type, e.g. "gp2" or "io1"."""

    root_device = attr.ib(type=str, default=None, validator=instance_of_str)
    """Root device node used with this image, e.g. "/dev/sda1"."""

    description = attr.ib(type=str, default=None, validator=instance_of_str)
    """A brief human-readable description of the image."""

    sriov_net_support = attr.ib(type=str, default=None, validator=instance_of_str)
    """"simple" if the image is SRIOV-enabled."""

    ena_support = attr.ib(
        type=bool, default=None, validator=optional(instance_of(bool))
    )
    """``True`` if the image supports Elastic Network Adapter (ENA)."""
    billing_codes = attr.ib(
        type=AmiBillingCodes,
        default=None,
        validator=optional(instance_of(AmiBillingCodes)),
    )
    """Billing codes associated with this image."""



import logging
from functools import partial

from pushsource._impl.compat import scandir

LOG = logging.getLogger("pushsource")


class TypeHandler(object):
    # Decorator for handling specific file directories (e.g. "FILES", "ISOS" etc)
    HANDLERS = {}

    def __init__(self, type_name):
        self.type_name = type_name

    def __call__(self, fn):
        TypeHandler.HANDLERS[self.type_name] = fn
        return fn


handles_type = TypeHandler


class StagedBaseMixin(object):
    # Helper mixin for staged classes handling specific pieces of content.
    _FILE_TYPES = {}

    def __init__(self, *args, **kwargs):
        super(StagedBaseMixin, self).__init__(*args, **kwargs)
        self._FILE_TYPES = self._FILE_TYPES.copy()
        for typename in TypeHandler.HANDLERS:
            fn = TypeHandler.HANDLERS[typename]
            bound_fn = partial(fn, self)
            self._FILE_TYPES[typename] = partial(
                self.__mixin_push_items, delegate=bound_fn
            )

    def __mixin_push_items(self, leafdir, metadata, delegate):
        out = []

        LOG.debug("Looking for files in %s", leafdir)

        for entry in scandir(leafdir.path):
            if entry.is_file():
                item = delegate(leafdir, metadata, entry)
                if item:
                    out.append(item)

        return out



from .base import PushItem
from .. import compat_attr as attr


@attr.s()
class ModuleMdPushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a modulemd stream.

    For push items of this type, the :meth:`~pushsource.PushItem.src` attribute
    refers to a file containing a YAML document stream. The stream is expected
    to contain one or more modulemd or modulemd-defaults documents.

    This library does not verify that the referenced file is a valid
    modulemd stream.
    """


@attr.s()
class ModuleMdSourcePushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a modulemd source/packager document.

    Similar to :class:`~pushsource.ModuleMdPushItem`, but refers to the source
    document, typically named ``modulemd.src.txt`` in koji.

    This library does not verify that the referenced file is a valid
    modulemd source document.
    """



from .base import PushItem
from .conv import optional_str
from .. import compat_attr as attr


@attr.s()
class RpmPushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a single RPM."""

    module_build = attr.ib(type=str, default=None, validator=optional_str)
    """NVR for the koji module build through which this push item was located, if any.

    For a modular RPM push item, there may be two koji builds in play:

    ``module_build`` (this attribute):
        NVR of a build which produced modulemd YAML files referencing this RPM
        (but the build itself does not contain RPMs).

        Example: `ghc-8.4-820200708061905.9edba152`_
        is a valid `module_build` for ``ghc-8.4.4-74.module_el8+12161+cf1bd7f2.x86_64.rpm``.

    :meth:`~pushsource.PushItem.build`:
        NVR of the build containing the RPM.

        Example: `ghc-8.4.4-74.module_el8+12161+cf1bd7f2`_
        is a valid build for ``ghc-8.4.4-74.module_el8+12161+cf1bd7f2.x86_64.rpm``.

    Note that the ``module_build`` itself isn't a property of the RPM but rather should
    be considered contextual information on how any given RPM was located. The same RPM
    may have different values of ``module_build`` when loaded from different contexts.

    .. _ghc-8.4-820200708061905.9edba152: https://koji.fedoraproject.org/koji/buildinfo?buildID=1767200
    .. _ghc-8.4.4-74.module_el8+12161+cf1bd7f2: https://koji.fedoraproject.org/koji/buildinfo?buildID=1767130
    """



import logging

from ...validator import Validator
from ... import compat_attr as attr

REQUIRED_VERSION = "0.2"
VALIDATOR = Validator("staged", ids=["relative_path"])
LOG = logging.getLogger("pushsource")


@attr.s()
class StagingFileMetadata(object):
    attributes = attr.ib(type=dict)
    filename = attr.ib(type=str)
    relative_path = attr.ib(type=str)
    sha256sum = attr.ib(type=str)
    version = attr.ib(type=str)


@attr.s()
class StagingMetadata(object):
    # A private class modelling the content of staging metadata files

    # Filename from which metadata was loaded.
    filename = attr.ib(type=str, default=None)

    # Metadata per file, keyed by relative path within staging area
    file_metadata = attr.ib(type=dict, default=attr.Factory(dict))

    def file_metadata_or_die(self, relative_path):
        # Return StagingFileMetadata corresponding to relative_path, or raise a fatal error
        # if not available.
        file_md = self.file_metadata.get(relative_path)
        if file_md is None:
            message = "No metadata available for %s" % relative_path
            if self.filename:
                message = "%s in %s" % (message, self.filename)
            raise ValueError(message)

        return file_md

    @classmethod
    def from_data(cls, data, filename="<unknown file>"):
        VALIDATOR.validate(data, filename)

        payload = data.get("payload") or {}
        files = payload.get("files") or []
        file_metadata = {}

        for entry in files:
            md = StagingFileMetadata(
                attributes=entry.get("attributes") or {},
                filename=entry.get("filename"),
                relative_path=entry["relative_path"],
                sha256sum=entry.get("sha256sum"),
                version=entry.get("version"),
            )
            if md.relative_path in file_metadata:
                raise ValueError(
                    "File %s listed twice in %s" % (md.relative_path, filename)
                )
            file_metadata[md.relative_path] = md

        return cls(filename=filename, file_metadata=file_metadata)


@attr.s()
class StagingLeafDir(object):
    # A private class modelling a leaf directory within a staging area.
    file_type = attr.ib(type=str)
    dest = attr.ib(type=str)
    path = attr.ib(type=str)
    topdir = attr.ib(type=str)



import yaml
import os


SCHEMA_PATH = os.path.dirname(__file__)


def get_schema(name):
    filename = "%s-schema.yaml" % name
    path = os.path.join(SCHEMA_PATH, filename)
    with open(path) as f:
        return yaml.load(f, Loader=yaml.SafeLoader)



"""Helpers to deal with "from" vs "from_" issue in ErratumPushItem.

The advisory model has a "from" field, which unfortunately clashes with
a Python keyword of the same name. This prevents directly declaring the
field as "from".

Despite the clash, it is desirable to have the attrs field named "from"
so that usage of this model does not have to be frequently accompanied
by some "from_" => "from" renaming (e.g. before uploading to Pulp).

It is possible to make this work by:

- in the ErratumPushItem class, declare the attribute as "from_"
- then, after the class is created, patch it to effectively rename it
  to "from"

This makes use of the __attrs_attrs__ attribute which is associated
with each attrs class. Though the name implies it is private, this is
a supported & documented method of extending the behavior of attrs,
see: https://www.attrs.org/en/stable/extending.html

A simpler trick may come to mind: just setattr(ErratumPushItem, 'from', attr.ib(...))
to work around the inability to write "from = attr.ib(...)".
Unfortunately, that does not work because the attrs library internally
tries to eval some code of the form "<attr_name> = ...", which will never
work if the name is a keyword.
"""


class AttrsRenamer(object):
    def __init__(self, delegate, attrs_old_to_new):
        """Helper to selectively rename certain attributes for __attrs_attrs__
        within a class.

        Arguments:
            delegate (tuple)
                Original instance of __attrs_attrs__, generated by attrs library.
            attrs_old_to_new (dict)
                A mapping from old to new attribute names.
        """
        self._delegate = delegate
        self._attrs_old_to_new = attrs_old_to_new

        self._attrs_by_name = {}
        for old_name, new_name in self._attrs_old_to_new.items():
            old_attr = getattr(delegate, old_name)

            # Make a new attribute, identical to old in all respects
            # except for the name.
            # (we do this dynamically to cope with differences between ancient and newer
            # versions of attrs library)
            attr_kwargs = {"name": new_name}
            for argname in [
                "default",
                "validator",
                "repr",
                "cmp",
                "hash",
                "init",
                "inherited",
            ]:
                if hasattr(old_attr, argname):
                    attr_kwargs[argname] = getattr(old_attr, argname)
            new_attr = old_attr.__class__(**attr_kwargs)

            self._attrs_by_name[new_name] = new_attr

    def __iter__(self):
        for elem in self._delegate:
            new_name = self._attrs_old_to_new.get(elem.name)
            if new_name:
                # Something we've renamed - yield ours
                yield self._attrs_by_name[new_name]
            else:
                # Something else - yield original
                yield elem

    def __getattr__(self, name):
        # If it's one of the new attribute names, just return it directly
        if name in self._attrs_by_name:
            return self._attrs_by_name[name]

        # If it's one of the old attribute names, make it not exist
        if name in self._attrs_old_to_new:
            raise AttributeError()

        # If it's anything else then just let the delegate handle it,
        # giving exactly normal behavior
        return getattr(self._delegate, name)


def fixup_attrs(cls):
    # Fix up the set of attributes on the class.
    #
    # This impacts dynamic use of the class e.g. via attr.fields, attr.asdict,
    # attr.evolve, ...
    cls.__attrs_attrs__ = AttrsRenamer(cls.__attrs_attrs__, {"from_": "from"})


def fixup_init(cls):
    # Fix up the constructor of the class to support both names
    # of the "from" attribute.
    #
    # As the underlying attribute is actually stored as "from_", we need to patch
    # the constructor to accept the "from" argument too and rename it for storage.
    #
    # If the caller provides both "from" and "from_", the former is preferred.
    cls_init = cls.__init__

    def new_init(*args, **kwargs):
        if "from" in kwargs:
            kwargs["from_"] = kwargs.pop("from") or kwargs.get("from_")
        return cls_init(*args, **kwargs)

    cls.__init__ = new_init


def fixup_props(cls):
    # Fix up properties of the class so that there is a "from" property which
    # aliases the underlying "from_".
    from_attr = property(lambda self: self.from_)
    setattr(cls, "from", from_attr)


def fixup_erratum_class(cls):
    # Do all the fixups to make "from" and "from_" aliases.
    fixup_attrs(cls)
    fixup_init(cls)
    fixup_props(cls)



import os
import threading
import logging
import itertools
import functools

import yaml
import json

from pushsource._impl.compat import scandir

from pushcollector import Collector
from more_executors import Executors

from ...source import Source
from ...helpers import list_argument, as_completed_with_timeout_reset

from .staged_utils import StagingMetadata, StagingLeafDir
from .staged_ami import StagedAmiMixin
from .staged_files import StagedFilesMixin
from .staged_errata import StagedErrataMixin
from .staged_compsxml import StagedCompsXmlMixin
from .staged_modulemd import StagedModuleMdMixin
from .staged_productid import StagedProductIdMixin
from .staged_rpm import StagedRpmMixin
from .staged_unsupported import StagedUnsupportedMixin

LOG = logging.getLogger("pushsource")
METADATA_FILES = ["staged.yaml", "staged.yml", "staged.json", "pub-mapfile.json"]
CACHE_LOCK = threading.RLock()


class StagedSource(
    Source,
    StagedAmiMixin,
    StagedFilesMixin,
    StagedErrataMixin,
    StagedCompsXmlMixin,
    StagedModuleMdMixin,
    StagedProductIdMixin,
    StagedRpmMixin,
    StagedUnsupportedMixin,
):
    """Uses a directory with a predefined layout (a "staging directory") as
    the source of push items."""

    # To keep code maintainable, this class doesn't directly support any file
    # types; separate files are expected to register themselves here.
    _FILE_TYPES = {}

    def __init__(self, url, threads=4, timeout=60 * 60):
        """Create a new source.

        Parameters:
            url (list[str])
                URL(s) of locally accessible staging directories, e.g.
                ``"/mnt/staging/my-content-for-push"``.

                These directories must follow the documented layout for staging areas.

            threads (int)
                Number of threads used for concurrent loading of files.

            timeout (int)
                Number of seconds after which an error is raised, if no progress is
                made during each step.

        """
        super(StagedSource, self).__init__()
        self._url = list_argument(url)
        self._threads = threads
        self._timeout = timeout

        # Note: this executor does not have a retry.
        # NFS already does a lot of its own retries.
        self._executor = (
            Executors.thread_pool(name="pushsource-staged", max_workers=threads)
            .with_timeout(timeout)
            .with_cancel_on_shutdown()
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._executor.shutdown()

    def __iter__(self):
        # Possible improvement would be to handle the separate dirs concurrently.
        # However, may not be worthwhile, as  in practice I'm not sure the ability
        # to pass more than a single staging directory is ever used (?)
        return itertools.chain(
            *[self._push_items_for_topdir(x) for x in self._url]
        ).__iter__()

    def _load_metadata(self, topdir):
        # Load the top-level metadata file in the staging directory, if any.
        for candidate in METADATA_FILES:
            metadata_file = os.path.join(topdir, candidate)
            if os.path.exists(metadata_file):
                break
        else:
            # no metadata file
            return StagingMetadata()

        basename = os.path.basename(metadata_file)

        with open(metadata_file, "rt") as f:
            content = f.read()

            # Save a copy of the file for later reference
            Collector.get().attach_file(basename, content).result()

            if metadata_file.endswith(".json"):
                metadata = json.loads(content)
            else:
                metadata = yaml.safe_load(content)

        return StagingMetadata.from_data(metadata, os.path.basename(metadata_file))

    def _push_items_for_leafdir(self, leafdir, metadata):
        LOG.debug("Scanning %s", leafdir.path)
        if not os.path.exists(leafdir.path):
            return []
        return self._FILE_TYPES[leafdir.file_type](leafdir=leafdir, metadata=metadata)

    def _push_items_for_topdir(self, topdir):
        LOG.info("Checking files in: %s", topdir)

        metadata = self._load_metadata(topdir)

        destdirs = []
        for entry in scandir(topdir):
            if entry.is_dir() and entry.name != "logs":
                destdirs.append(entry.path)

        if not destdirs:
            LOG.warning("%s has no destination directories", topdir)
            # If there's no directories AND no metadata file, the caller most likely
            # provided an incorrect path. (If the caller expects to provide empty
            # staging areas sometimes, they can signal this by including empty metadata.)
            if not metadata.filename:
                raise IOError("%s does not appear to be a staging directory" % topdir)
            return

        # All directories which may contain files to be processed
        all_leaf_dirs = []
        for destdir in destdirs:
            for file_type in self._FILE_TYPES:
                path = os.path.join(destdir, file_type)
                all_leaf_dirs.append(
                    StagingLeafDir(
                        dest=os.path.basename(destdir),
                        file_type=file_type,
                        path=path,
                        topdir=topdir,
                    )
                )

        process_dir = functools.partial(self._push_items_for_leafdir, metadata=metadata)
        pushitems_fs = [
            self._executor.submit(process_dir, leafdir) for leafdir in all_leaf_dirs
        ]

        completed_fs = as_completed_with_timeout_reset(
            pushitems_fs, timeout=self._timeout
        )
        for f in completed_fs:
            for pushitem in f.result():
                yield pushitem


Source._register_backend_builtin("staged", StagedSource)



from frozenlist2 import frozenlist

from .base import PushItem
from .. import compat_attr as attr
from .conv import (
    in_,
    int2str,
    md5str,
    sha1str,
    sha256str,
    instance_of,
    instance_of_str,
    optional,
    optional_str,
)
from .erratum_fixup import fixup_erratum_class


@attr.s()
class ErratumReference(object):
    """A reference within a :meth:`~ErratumPushItem.references` list."""

    href = attr.ib(type=str, validator=instance_of_str)
    """A URL."""

    id = attr.ib(type=str, converter=int2str, validator=optional_str)
    """A short ID for the reference, unique within this erratum."""

    title = attr.ib(type=str, validator=optional_str)
    """A title for the reference; analogous to the 'title' attribute
    in HTML.
    """

    type = attr.ib(type=str, default="other", validator=instance_of_str)
    """Type of reference. This defines the expected target of the URL
    and includes at least:

    - "self": reference to a page for this advisory
    - "bugzilla": reference to a bug
    - "other": any other kind of reference
    """

    @classmethod
    def _from_data(cls, data):
        # Convert from raw list/dict as provided in ET APIs into model
        if isinstance(data, list):
            return [cls._from_data(elem) for elem in data]

        return cls(
            href=data["href"], id=data["id"], title=data["title"], type=data["type"]
        )


@attr.s()
class ErratumModule(object):
    """A module entry within a :meth:`~ErratumPushItem.pkglist`."""

    name = attr.ib(type=str, validator=instance_of_str)
    """Module name."""

    stream = attr.ib(type=str, validator=instance_of_str)
    """Module stream."""

    version = attr.ib(type=str, validator=instance_of_str)
    """Module version."""

    context = attr.ib(type=str, validator=instance_of_str)
    """Module context."""

    arch = attr.ib(type=str, validator=instance_of_str)
    """Module architecture."""

    def __str__(self):
        return ":".join([self.name, self.stream, self.version, self.context, self.arch])

    @classmethod
    def _from_data(cls, data):
        if not data:
            return None
        return cls(
            arch=data["arch"],
            context=data["context"],
            name=data["name"],
            stream=data["stream"],
            version=data["version"],
        )


@attr.s()
class ErratumPackage(object):
    """A package (RPM) entry within a :meth:`~ErratumPushItem.pkglist`."""

    arch = attr.ib(type=str, validator=instance_of_str)
    """RPM architecture."""

    filename = attr.ib(type=str, validator=instance_of_str)
    """RPM filename (basename)."""

    epoch = attr.ib(type=str, converter=int2str, validator=instance_of_str)
    """RPM epoch."""

    name = attr.ib(type=str, validator=instance_of_str)
    """RPM name (e.g. "bash-4.0.1-1.el7.x86_64.rpm" name is "bash")"""

    version = attr.ib(type=str, validator=instance_of_str)
    """RPM version (e.g. "bash-4.0.1-1.el7.x86_64.rpm" version is "4.0.1")"""

    release = attr.ib(type=str, validator=instance_of_str)
    """RPM release (e.g. "bash-4.0.1-1.el7.x86_64.rpm" version is "1.el7")"""

    src = attr.ib(type=str, validator=instance_of_str)
    """Filename of the source RPM from which this RPM was built; equal to
    :meth:`filename` for the source RPM itself.
    """

    reboot_suggested = attr.ib(type=bool, default=False, validator=instance_of(bool))
    """True if rebooting host machine is recommended after installing this package."""

    md5sum = attr.ib(type=str, default=None, converter=md5str)
    """MD5 checksum of this RPM in hex string form, if available."""

    sha1sum = attr.ib(type=str, default=None, converter=sha1str)
    """SHA1 checksum of this RPM in hex string form, if available."""

    sha256sum = attr.ib(type=str, default=None, converter=sha256str)
    """SHA256 checksum of this RPM in hex string form, if available."""


@attr.s()
class ErratumPackageCollection(object):
    """A collection of packages found within an :meth:`~ErratumPushItem.pkglist`.

    A non-modular advisory typically contains only a single collection, while modular
    advisories typically contain one collection per module.
    """

    name = attr.ib(type=str, default="", validator=instance_of_str)
    """A name for this collection. The collection name has no specific meaning,
    but must be unique within an advisory.
    """

    packages = attr.ib(
        type=list, default=attr.Factory(frozenlist), converter=frozenlist
    )
    """List of packages within this collection.

    :type: list[ErratumPackage]
    """

    short = attr.ib(type=str, default="", validator=instance_of_str)
    """An alternative name for this collection. In practice, this field
    is typically blank.
    """

    module = attr.ib(
        type=ErratumModule, default=None, validator=optional(instance_of(ErratumModule))
    )
    """An :class:`~ErratumModule` defining the module this entry is associated
    with, if any.
    """

    @classmethod
    def _from_data(cls, data):
        # Convert from raw list/dict as provided in ET APIs into model.
        # Data is expected to be 'pkglist' object.
        if isinstance(data, list):
            out = [cls._from_data(elem) for elem in data]
            # only include non-empty collections
            return [elem for elem in out if elem.packages]

        packages = []
        for raw_pkg in data.get("packages") or []:
            # parse the odd 'sum' structure, which is a list of form:
            # [<algo>, <hexdigest>, <algo>, <hexdigest>, ...]
            sums = {}
            raw_sum = raw_pkg.pop("sum", [])
            while raw_sum:
                sums[raw_sum[0]] = raw_sum[1]
                raw_sum = raw_sum[2:]

            packages.append(
                ErratumPackage(
                    arch=raw_pkg["arch"],
                    epoch=raw_pkg["epoch"],
                    filename=raw_pkg["filename"],
                    name=raw_pkg["name"],
                    version=raw_pkg["version"],
                    release=raw_pkg["release"],
                    src=raw_pkg["src"],
                    reboot_suggested=raw_pkg.get("reboot_suggested") or False,
                    md5sum=sums.get("md5"),
                    sha1sum=sums.get("sha1"),
                    sha256sum=sums.get("sha256"),
                )
            )

        return cls(
            name=data.get("name") or "",
            short=data.get("short") or "",
            packages=packages,
            module=ErratumModule._from_data(data.get("module")),
        )


def errata_type_converter(value):
    # For 'type' field, ancient versions of Errata Tool produced values
    # like RHBA, RHEA, RHSA, so these can be found in the wild; but all
    # other sources (including our own model) prefer strings like "bugfix".
    # We'll convert up from legacy values automatically.
    typemap = {"RHBA": "bugfix", "RHEA": "enhancement", "RHSA": "security"}
    return typemap.get(value, value)


@attr.s()
class ErratumPushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a single erratum (also known as "advisory").

    Note that many fields on erratum objects which appear to be numeric
    are instead represented as strings ('release' and 'pushcount' being two
    examples).
    """

    type = attr.ib(
        type=str,
        default="bugfix",
        validator=in_(["bugfix", "security", "enhancement"]),
        converter=errata_type_converter,
    )
    """'bugfix', 'security' or 'enhancement'."""

    release = attr.ib(
        type=str, default="0", converter=int2str, validator=instance_of_str
    )
    """Release number. Typically an integer-string, initially '0'."""

    status = attr.ib(type=str, default="final", validator=instance_of_str)
    """Status, typically 'final'."""

    pushcount = attr.ib(
        type=str, default="1", converter=int2str, validator=instance_of_str
    )
    """Number of times advisory has been revised and published (starting at '1')."""

    reboot_suggested = attr.ib(type=bool, default=False, validator=instance_of(bool))
    """True if rebooting host machine is recommended after installing this advisory.

    .. warning::
        The intended usage of this field is unclear.

        In practice, tools such as yum are instead consuming the ``reboot_suggested`` field
        from :class:`ErratumPackage`.
    """

    references = attr.ib(
        type=list, default=attr.Factory(frozenlist), converter=frozenlist
    )
    """A list of references associated with the advisory.

    :type: list[ErratumReference]
    """

    pkglist = attr.ib(type=list, default=attr.Factory(frozenlist), converter=frozenlist)
    """A list of package collections associated with the advisory.

    :type: list[ErratumPackageCollection]
    """

    from_ = attr.ib(type=str, default=None, validator=optional_str)
    """Contact email address for the owner of the advisory.

    Note that the canonical name for this attribute is ``from``. As this clashes
    with a Python keyword, in most contexts the attribute is also available as an
    alias, ``from_``. Where possible, the canonical name ``from`` should be preferred.

    :type: str
    """

    rights = attr.ib(type=str, default=None, validator=optional_str)
    """Copyright message."""

    title = attr.ib(type=str, default=None, validator=optional_str)
    """Title of the advisory (e.g. 'bash bugfix and enhancement')."""

    description = attr.ib(type=str, default=None, validator=optional_str)
    """Full human-readable description of the advisory, usually multiple lines."""

    version = attr.ib(
        type=str, default="1", converter=int2str, validator=instance_of_str
    )
    """Advisory version. Starts counting at "1", and some systems require updating
    the version whenever an advisory is modified.
    """

    updated = attr.ib(type=str, default=None, validator=optional_str)
    """Timestamp of the last update to this advisory.

    Typically of the form '2019-12-31 06:54:41 UTC', but this is not enforced.
    """

    issued = attr.ib(type=str, default=None, validator=optional_str)
    """Timestamp of the initial release of this advisory.

    Uses the same format as :meth:`updated`.
    """

    severity = attr.ib(type=str, default=None, validator=optional_str)
    """Severity of the advisory, e.g. "low", "moderate", "important" or "critical"."""

    summary = attr.ib(type=str, default=None, validator=optional_str)
    """Typically a single sentence briefly describing the advisory."""

    solution = attr.ib(type=str, default=None, validator=optional_str)
    """Text explaining how to apply the advisory."""

    content_types = attr.ib(
        type=list, default=attr.Factory(frozenlist), converter=frozenlist
    )
    """A list of content types associated with the advisory.

    For example, "docker" may be found in this list if the advisory deals
    with container images.

    :type: list[str]
    """

    def __str__(self):
        return "%s: %s" % (self.name, self.title or "<untitled advisory>")

    @classmethod
    def _from_data(cls, data):
        # Helper to load an instance from the raw dicts used in various places:
        #
        # - json/yaml files in staging area
        # - output of get_advisory_cdn_metadata method in ET
        # - erratum unit metadata in Pulp2's API
        #

        kwargs = {}

        # Fill in the base push item attributes first.
        kwargs["name"] = data["id"]
        kwargs["state"] = "PENDING"

        # Now the erratum-specific fields.
        # Many are accepted verbatim from raw input; these are mandatory
        # (i.e. we'll crash if the field is missing from input).
        for field in [
            "type",
            "release",
            "status",
            "pushcount",
            "reboot_suggested",
            "rights",
            "title",
            "from",
            "description",
            "version",
            "updated",
            "issued",
            "severity",
            "summary",
            "solution",
        ]:
            kwargs[field] = data[field]

        # If the metadata has a cdn_repo field, this sets the destinations for the
        # push item; used for text-only errata.
        if data.get("cdn_repo"):
            kwargs["dest"] = data["cdn_repo"]

        # If there are content type hints, copy those while dropping the
        # pulp-specific terminology
        pulp_user_metadata = data.get("pulp_user_metadata") or {}
        if pulp_user_metadata.get("content_types"):
            kwargs["content_types"] = pulp_user_metadata["content_types"]

        kwargs["references"] = ErratumReference._from_data(data.get("references") or [])

        kwargs["pkglist"] = ErratumPackageCollection._from_data(
            data.get("pkglist") or []
        )

        return cls(**kwargs)


# Apply from vs from_ workarounds.
fixup_erratum_class(ErratumPushItem)



import logging

try:
    from kobo import rpmlib
except Exception as ex:  # pragma: no cover, pylint: disable=broad-except
    # If kobo.rpmlib is unavailable, let's not immediately crash.
    # We will hold this exception and re-raise it only if there's an
    # attempt to use the related functionality.
    from .. import broken_rpmlib as rpmlib

    rpmlib.CAUSE = ex

from ...model import RpmPushItem
from .staged_base import StagedBaseMixin, handles_type

LOG = logging.getLogger("pushsource")


class StagedRpmMixin(StagedBaseMixin):
    @handles_type("RPMS")
    @handles_type("SRPMS")
    def __push_item(self, leafdir, _, entry):
        if not entry.name.endswith(".rpm"):
            # Old code accepted (ignored) non-RPMs in RPMs dir.
            # It's not clear how widespread this is, so we'd better tolerate it,
            # but we don't like it, so let's warn.
            LOG.warning("Unexpected non-RPM %s (ignored)", entry.path)
            return None

        header = rpmlib.get_rpm_header(entry.path)
        key_id = rpmlib.get_keys_from_header(header)

        return RpmPushItem(
            name=entry.name,
            src=entry.path,
            signing_key=key_id,
            origin=leafdir.topdir,
            dest=[leafdir.dest],
        )



from .base import PushItem, KojiBuildInfo
from .erratum import (
    ErratumPushItem,
    ErratumReference,
    ErratumModule,
    ErratumPackage,
    ErratumPackageCollection,
)
from .rpm import RpmPushItem
from .file import FilePushItem
from .container import (
    ContainerImagePushItem,
    ContainerImagePullInfo,
    ContainerImagePullSpec,
    ContainerImageDigestPullSpec,
    ContainerImageTagPullSpec,
    SourceContainerImagePushItem,
    OperatorManifestPushItem,
)
from .modulemd import ModuleMdPushItem, ModuleMdSourcePushItem
from .comps import CompsXmlPushItem
from .productid import ProductIdPushItem
from .ami import AmiPushItem, AmiRelease, AmiBillingCodes



from .base import PushItem
from .. import compat_attr as attr


@attr.s()
class ProductIdPushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a product ID certificate.

    For push items of this type, the :meth:`~pushsource.PushItem.src` attribute
    refers to a file containing a PEM certificate identifying a product.

    This library does not verify that the referenced file is a valid
    certificate.
    """



from .base import PushItem
from .. import compat_attr as attr


@attr.s()
class CompsXmlPushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a comps XML file.

    For push items of this type, the :meth:`~pushsource.PushItem.src`
    attribute refers to a file containing a comps XML file (package
    group definitions) in the format used by yum & dnf.

    This library does not verify that the referenced file is valid.
    """



import re
import logging

from ..source import Source
from ..model import (
    ContainerImagePushItem,
    SourceContainerImagePushItem,
    ContainerImagePullInfo,
    ContainerImageTagPullSpec,
    ContainerImageDigestPullSpec,
)
from ..utils.containers import (
    get_manifest,
    inspect,
    MEDIATYPE_SCHEMA2_V2,
    MEDIATYPE_SCHEMA2_V1,
    MEDIATYPE_SCHEMA2_V1_SIGNED,
    MEDIATYPE_SCHEMA2_V2_LIST,
)

from ..helpers import list_argument

LOG = logging.getLogger("pushsource")

IMAGE_URI_REGEX = re.compile("https://([^:@]*):(.+)(([^:]:)+)*")


class RegistrySource(Source):
    """Uses a container image registry as a source of push items."""

    def __init__(self, image, dest=None, dest_signing_key=None):
        """Create a new source.

        Parameters:
            image (str, list[str])
                Pull spec(s) of container images, as a list or a comma-separated string.

                Each pull spec must include a hostname and a tag; referencing images by digest
                is currently not supported.

                ``registry.access.redhat.com/ubi8/ubi:8.4-211`` is an example of a valid pull
                spec.

            dest (str, list[str])
                If provided, this value will be used to populate :meth:`~pushsource.PushItem.dest`
                on generated push items.

            dest_signing_key (str, list[str])
                If provided, this value will be used to populate
                :meth:`~pushsource.ContainerImagePushItem.dest_signing_key` on generated
                push items.

                Note that, as each item holds only a single ``dest_signing_key``, using this
                argument can affect the number of generated push items. For example,
                providing two keys would produce double the amount of push items as providing
                a single key.
        """
        self._images = ["https://%s" % x for x in list_argument(image)]
        if dest:
            self._repos = list_argument(dest)
        else:
            self._repos = []
        self._signing_keys = list_argument(dest_signing_key)
        self._inspected = {}
        self._manifests = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def _push_item_from_registry_uri(self, uri, signing_key):
        if not IMAGE_URI_REGEX.match(uri):
            raise ValueError(
                "Invalid item uri: %s. Please use format: %s"
                % (uri, IMAGE_URI_REGEX.pattern)
            )

        schema, rest = uri.split("://")
        host, rest = rest.split("/", 1)
        repo, src_tag = rest.split(":", 1)
        source_uri = "%s/%s:%s" % (host, repo, src_tag)
        if source_uri not in self._inspected:
            self._inspected[source_uri] = inspect(
                "%s://%s" % (schema, host), repo, src_tag
            )
        if self._inspected[source_uri].get("source"):
            klass = SourceContainerImagePushItem
        else:
            klass = ContainerImagePushItem

        for mtype in [
            MEDIATYPE_SCHEMA2_V1,
            MEDIATYPE_SCHEMA2_V2,
            MEDIATYPE_SCHEMA2_V2_LIST,
        ]:
            if source_uri not in self._manifests:
                self._manifests[source_uri] = {}
            if mtype not in self._manifests[source_uri]:
                manifest_details = get_manifest(
                    "%s://%s" % (schema, host),
                    repo,
                    src_tag,
                    manifest_types=[mtype],
                )
                self._manifests[source_uri][mtype] = manifest_details

            manifest_details = self._manifests[source_uri][mtype]
            content_type, _, _ = manifest_details
            # tolerate text/plain as MEDIATYPE_SCHEMA2_V1 (due to wrong configuration in CDN)
            if content_type not in [
                MEDIATYPE_SCHEMA2_V2,
                MEDIATYPE_SCHEMA2_V1,
                MEDIATYPE_SCHEMA2_V1_SIGNED,
                MEDIATYPE_SCHEMA2_V2_LIST,
                "text/plain",
            ]:
                raise ValueError("Unsupported manifest type: %s" % content_type)

        pull_info = ContainerImagePullInfo(
            digest_specs=[
                ContainerImageDigestPullSpec(
                    registry=host,
                    repository=repo,
                    digest=manifest[1],
                    # replace text/plain with correct value
                    media_type=manifest[0].replace("text/plain", MEDIATYPE_SCHEMA2_V1),
                )
                for manifest in self._manifests[source_uri].values()
            ],
            media_types=[
                manifest[0].replace("text/plain", MEDIATYPE_SCHEMA2_V1)
                for manifest in self._manifests[source_uri].values()
            ],
            tag_specs=[
                ContainerImageTagPullSpec(
                    registry=host,
                    repository=repo,
                    tag=src_tag,
                    media_types=[
                        manifest[0].replace("text/plain", MEDIATYPE_SCHEMA2_V1)
                        for manifest in self._manifests[source_uri].values()
                    ],
                )
            ],
        )
        return klass(
            name=source_uri,
            dest=self._repos,
            dest_signing_key=signing_key,
            src=source_uri,
            source_tags=[src_tag],
            labels=self._inspected.get(source_uri, {}).get("config").get("Labels", {}),
            arch=(self._inspected.get(source_uri, {}).get("config", {}) or {})
            .get("Labels", {})
            .get("Labels", {})
            .get("architecture"),
            pull_info=pull_info,
        )

    def __iter__(self):
        for key in self._signing_keys or [None]:
            for uri in self._images:
                yield self._push_item_from_registry_uri(uri, key)


Source._register_backend_builtin("registry", RegistrySource)



import six

# A replacement for kobo.rpmlib to give a decent error message
# when rpm bindings are not available.

# importer is expected to put a more specific cause here.
CAUSE = RuntimeError("unknown error")

MESSAGE = "kobo.rpmlib is not available (consider 'pip install rpm-py-installer')"


def not_available(*_args, **_kwargs):
    six.raise_from(RuntimeError(MESSAGE), CAUSE)


get_rpm_header = not_available
get_keys_from_header = not_available



import logging

from .staged_base import StagedBaseMixin, handles_type

LOG = logging.getLogger("pushsource")


class StagedUnsupportedMixin(StagedBaseMixin):
    @handles_type("DOCKER")
    @handles_type("CHANNEL_DUMPS")
    def __push_item(self, _leafdir, _metadata, entry):
        LOG.error("Unsupported content found: %s", entry.path)
        return None



import sys

import attr

# Wrappers for attr module to deal with some incompatibilities between versions


def s():
    kwargs = {"frozen": True}
    if sys.version_info >= (3,):
        kwargs["kw_only"] = True
    return attr.s(**kwargs)


ib = attr.ib
evolve = attr.evolve
validators = attr.validators
Factory = attr.Factory



# Utilities for dealing with modulemd files from backends.
# None of this is public API.
import yaml

from .. import compat_attr as attr


@attr.s()
class Module(object):
    MOD_NSVCA_FMT = "{name}:{stream}:{version}:{context}:{arch}"

    name = attr.ib()
    stream = attr.ib()
    version = attr.ib()
    context = attr.ib()
    arch = attr.ib()

    @property
    def nsvca(self):
        return self.MOD_NSVCA_FMT.format(
            name=self.name,
            stream=self.stream,
            version=self.version,
            context=self.context,
            arch=self.arch,
        )

    @classmethod
    def from_file(cls, fname):
        # Obtain a Module instance from a YAML file.
        # Raises if file doesn't contain a module or has multiple documents.
        with open(fname) as f:
            parsed = yaml.load(f, Loader=yaml.BaseLoader)
        data = parsed["data"]

        return cls(
            name=data["name"],
            stream=data["stream"],
            version=data["version"],
            context=data["context"],
            arch=data["arch"],
        )



from .base import PushItem
from .. import compat_attr as attr
from .conv import optional_str


@attr.s()
class FilePushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a single generic file."""

    description = attr.ib(type=str, default=None, validator=optional_str)
    """A human-readable brief description of the file."""

    version = attr.ib(type=str, default=None, validator=optional_str)
    """A version string associated with the file.

    This string is intended for display purposes only.
    It may denote a product version associated with this file.
    For example, a push item for ``oc-4.2.33-linux.tar.gz`` may
    use a version of ``"4.2.33"`` to denote that the file relates
    to OpenShift version 4.2.33.
    """



# Converters and validators for use with model classes.
import datetime
import logging
from functools import partial
import re

import six
from frozenlist2 import frozenlist
from dateutil import tz

from .. import compat_attr as attr


LOG = logging.getLogger("pushsource")
HEX_PATTERN = re.compile(r"^[0-9a-f]+$")


def sloppylist(value, elem_converter=None):
    """Accept real lists or comma-separated values, and output a frozen list.

    Optionally use elem_converter to convert each list element.
    """
    if isinstance(value, six.string_types):
        value = value.split(",")
    if elem_converter:
        value = [elem_converter(elem) for elem in value]
    return frozenlist(value)


sloppyintlist = partial(sloppylist, elem_converter=int)


def timestamp(value):
    if isinstance(value, six.string_types):
        for fmt in (
            "%Y-%m-%dT%H:%M:%SZ",
            "%Y-%m-%dT%H:%M:%S",
            "%Y-%m-%dT%H:%MZ",
            "%Y-%m-%dT%H:%M",
        ):
            try:
                value = datetime.datetime.strptime(value, fmt)
                break
            except ValueError:
                LOG.debug("can't parse %s using %s", value, fmt, exc_info=1)
        else:
            raise ValueError("can't parse %s as a timestamp" % repr(value))

    # If it's not timezone-aware, force UTC
    if not value.tzinfo:
        value = value.replace(tzinfo=tz.tzutc())

    return value


def datestr(value):
    if isinstance(value, six.string_types):
        for fmt in ("%Y%m%d", "%Y-%m-%d"):
            try:
                value = datetime.datetime.strptime(value, fmt)
                value = datetime.date(value.year, value.month, value.day)
                break
            except ValueError:
                LOG.debug("can't parse %s using %s", value, fmt, exc_info=1)
        else:
            raise ValueError("can't parse %s as a date" % repr(value))

    return value


def hexstr(length, value):
    if value is None:
        return value

    msg = "can't parse %s as a hex string" % repr(value)

    if not isinstance(value, six.string_types):
        raise TypeError(msg)

    if len(value) != length:
        raise ValueError("wrong length of %s (expected: %s)" % (repr(value), length))

    value = value.lower()
    if not HEX_PATTERN.match(value):
        raise ValueError(msg)

    return value


md5str = partial(hexstr, 32)
sha1str = partial(hexstr, 40)
sha256str = partial(hexstr, 64)


def archstr(value):
    """Convert a handful of arch aliases into canonical form for consistent comparisons."""
    if value == "amd64":
        return "x86_64"
    if value == "SRPM":
        return "src"
    return value


def upper_if_str(value):
    if isinstance(value, six.string_types):
        return value.upper()
    return value


def int2str(value):
    # If value is an integer, convert it to string instead.
    # Mostly used for "numeric string" fields on errata, such as pushcount.
    if isinstance(value, int):
        return str(value)
    return value


in_ = attr.validators.in_
instance_of = attr.validators.instance_of
instance_of_str = instance_of(six.string_types)
optional = attr.validators.optional
optional_str = optional(instance_of_str)



from concurrent.futures import wait, FIRST_COMPLETED

import six

# in py2 TimeoutError is not built-in exception
if six.PY2:  # pragma: no cover
    from concurrent.futures import TimeoutError  # pylint: disable=redefined-builtin

from six.moves.urllib.parse import urlparse, ParseResult


def list_argument(value, retain_none=False):
    """Convert an argument into a list:

    - if the value is already a list, it's returned verbatim
    - if the value is None and retain_none is False, an empty list is returned
    - if the value is None and retain_none is True, None is returned
    - if the value is a string, it's split on ","
    - otherwise, the value is wrapped in a list

    This function is intended for use in :class:`~pushsource.Source` constructors
    to ease the processing of arguments passed via URL.

    Parameters:
        value
            Any value.

    Returns:
        list
            ``value`` converted to or wrapped into a list.
        None
            if input was None and retain_none is True.
    """
    if isinstance(value, list):
        return value
    if value is None and retain_none:
        return value
    if not value:
        return []
    if isinstance(value, six.string_types):
        return value.split(",")
    return [value]


def try_int(value):
    """Convert strings into integers where possible.

    This function is intended for use in :class:`~pushsource.Source` constructors
    to ease the processing of arguments passed via URL.

    Parameters:
        value
            Any value.

    Returns:
        int
            ``value`` converted to an integer, if it was a string representation
            of an integer.
        other
            ``value`` unmodified, in any other case.
    """
    if not isinstance(value, six.string_types):
        return value
    try:
        return int(value)
    except ValueError:
        return value


def try_bool(value):
    """Convert values into booleans where possible.

    This function is intended for use in :class:`~pushsource.Source` constructors
    to ease the processing of arguments passed via URL (which always arrive as a string).

    Parameters:
        value
            Any value.

    Returns:
        bool
            ``value`` converted to a boolean, if it was a string representation
            of a boolean.
        other
            ``value`` unmodified, in any other case.
    """
    if not isinstance(value, six.string_types):
        return value

    value = value.lower()
    if value in ["1", "true", "yes"]:
        return True
    elif value in ["", "0", "false", "no"]:
        return False
    raise ValueError("Expected a boolean, got %s" % repr(value))


def force_https(url):
    """Force ``url`` to use https as its scheme.

    Parameters:
        url: str
            A string representation of a URL.

    Returns:
        str
            ``url`` with its scheme replaced with https.

    Throws:
        Any potential error thrown by the urlparse function during parsing.
    """
    parsed = urlparse(url)
    with_https = ParseResult("https", *parsed[1:])
    return with_https.geturl()


def as_completed_with_timeout_reset(futures, timeout):
    """
    Yields any finished future within time alloted by timeout, raises TimeoutError otherwise.
    Futures are allowed to continue over timeout, if any processing in caller is slow.
    Timeout is not applied to futures themselves but rather to wait() call, which means
    that this function allows to yield all quicker futures and tries to wait for slow ones.

    Parameters:
        futures: List[Future]
            list of futures

        timeout: int or float
            Timeout used for wait() calls

    Returns:
        Future
            finished Future object

    Throws:
        TimeoutError, if waiting for some Future to finish in wait() call takes more
        time than given timeout.
    """
    total_futures = len(futures)
    pending = futures
    while pending:
        finished, pending = wait(pending, timeout=timeout, return_when=FIRST_COMPLETED)
        # no future finished in timeout given, raise TimeoutError
        if not finished:
            raise TimeoutError(
                "%d (of %d) futures unfinished" % (len(pending), total_futures)
            )

        for fs in finished:
            yield fs



import re

import six
from frozenlist2 import frozenlist

from .base import PushItem
from .. import compat_attr as attr
from ..compat import frozendict
from .conv import instance_of, optional_str


MEDIA_TYPE_ORDER = {
    "application/vnd.docker.distribution.manifest.list.v2+json": 30,
    "application/vnd.docker.distribution.manifest.v2+json": 20,
    "application/vnd.docker.distribution.manifest.v1+json": 10,
}


@attr.s()
class ContainerImagePullSpec(object):
    """A container image pull spec in parsed form.

    Pull spec refers to a ``"NAME[:TAG|@DIGEST]"`` string used to pull
    a container image. This class provides a pull spec both in raw
    string form and parsed into various components, along with information
    on the available media type.
    """

    registry = attr.ib(type=str, validator=instance_of(six.string_types))
    """Registry component of pull spec.

    Example: in ``"registry.access.redhat.com/ubi8/ubi-minimal:latest"``,
    the registry is ``"registry.access.redhat.com"``.
    """

    repository = attr.ib(type=str, validator=instance_of(six.string_types))
    """Repository component of pull spec.

    Example: in ``"registry.access.redhat.com/ubi8/ubi-minimal:latest"``,
    the repository is ``"ubi8/ubi-minimal"``.
    """

    @classmethod
    def _from_str(cls, pull_spec):
        # Regex explanation
        # <registry>: Get everything not "/" until the first "/"
        # <repository>: Get everything not "@" or ":"
        # <type>: Either '@' or ':'
        # <digest_or_tag>: Get the rest of the characters until end of word
        url_pattern = (
            r"^(?P<registry>[^\/]+)\/(?P<repository>[^:@]+)"
            r"(?P<type>[:@])(?P<digest_or_tag>.+)$"
        )
        match = re.match(url_pattern, pull_spec)
        if not match:
            raise ValueError("Invalid container image pull spec '%s'" % pull_spec)

        out = {
            "registry": match.group("registry"),
            "repository": match.group("repository"),
        }
        if match.group("type") == "@":
            out["digest"] = match.group("digest_or_tag")
            klass = ContainerImageDigestPullSpec
        else:
            out["tag"] = match.group("digest_or_tag")
            klass = ContainerImageTagPullSpec

        return klass(**out)


@attr.s()
class ContainerImageTagPullSpec(ContainerImagePullSpec):
    """A container image pull spec using a tag."""

    tag = attr.ib(type=str, validator=instance_of(six.string_types))
    """Tag component of pull spec.

    Example: in ``"registry.access.redhat.com/ubi8/ubi-minimal:latest"``,
    the tag is ``"latest"``.
    """

    media_types = attr.ib(
        type=list, default=attr.Factory(frozenlist), converter=frozenlist
    )
    """Media type(s) expected to be reachable via this tag, or empty
    if unknown.

    Generally, the media types are only known for manifest lists obtained
    from a koji build.

    Example: ``["application/vnd.docker.distribution.manifest.list.v2+json"]``

    :type: List[str]
    """

    def __str__(self):
        return "".join([self.registry, "/", self.repository, ":", self.tag])


@attr.s()
class ContainerImageDigestPullSpec(ContainerImagePullSpec):
    """A container image pull spec using a manifest digest."""

    digest = attr.ib(type=str, validator=instance_of(six.string_types))
    """Digest component of pull spec.

    Example: in ``"registry.access.redhat.com/ubi8/ubi-minimal@sha256:0ccb9988abbc72d383258d58a7f519a10b637d472f28fbca6eb5fab79ba82a6b"``,
    the digest is ``"sha256:0ccb9988abbc72d383258d58a7f519a10b637d472f28fbca6eb5fab79ba82a6b"``.
    """

    media_type = attr.ib(type=str, default=None, validator=optional_str)
    """Media type of the manifest obtained by this pull spec, if known.

    Generally, the media type is known for all digest pull specs obtained
    via koji. It can be checked to determine whether a pull spec refers to
    a manifest list or an image manifest.

    Example: ``"application/vnd.docker.distribution.manifest.v2+json"``
    """

    def __str__(self):
        return "".join([self.registry, "/", self.repository, "@", self.digest])


def specs_converter(specs, expected_class):
    # a converter for pull specs, ensures every spec is an
    # instance of the expected class while de-duplicating
    out = []

    out_strs = set()
    for spec in specs:
        if not isinstance(spec, expected_class):
            raise TypeError("Expected %s, got: %s" % (expected_class, repr(spec)))

        # de-duplicate specs
        if str(spec) not in out_strs:
            out_strs.add(str(spec))
            out.append(spec)

    return frozenlist(out)


def tag_specs_converter(specs):
    out = specs_converter(specs, ContainerImageTagPullSpec)

    # Require at least one
    if not out:
        raise ValueError("A container image must have at least one pull spec")

    return out


def digest_specs_converter(specs):
    # Do the generic handling
    out = specs_converter(specs, ContainerImageDigestPullSpec)

    # Then also sort them by preferred media types
    out = sorted(out, key=lambda spec: -MEDIA_TYPE_ORDER.get(spec.media_type, 0))

    return frozenlist(out)


@attr.s()
class ContainerImagePullInfo(object):
    """Information needed to access a container image from a registry."""

    tag_specs = attr.ib(type=list, converter=tag_specs_converter)
    """Pull specs to access this image by tag.

    This may include specs which refer to a *manifest list* referencing this
    image, and also specs which refer directly to the *image manifest* for
    this image.

    This field always contains at least one pull spec. If you don't care
    which tag you use, it's reasonable to use ``item.pull_info.tag_specs[0]``.

    Note that, as all tags are mutable, it's possible for this information to
    be outdated. It is recommended to pull by digest or at least
    cross-reference with digests.

    :type: List[ContainerImageTagPullSpec]
    """

    digest_specs = attr.ib(type=list, converter=digest_specs_converter)
    """Pull specs to access this image by manifest digest.

    This may include specs which refer to a *manifest list* referencing this
    image, and also specs which refer directly to the *image manifest* for
    this image.

    These specs are always ordered from most to least preferred manifest type,
    which means:

    - manifest list if available, else:
    - schema2 image manifest if available, else:
    - schema1 image manifest

    :type: List[ContainerImageDigestPullSpec]
    """

    media_types = attr.ib(type=list, converter=frozenlist)
    """All media types for which a manifest is reachable in :meth:`digest_specs`.

    For information on these types see the `image manifest spec`_.

    :type: List[str]

    .. _image manifest spec: http://web.archive.org/web/20210614172014/https://docs.docker.com/registry/spec/manifest-v2-2/#media-types
    """

    @media_types.default
    def _default_media_types(self):
        return frozenlist(
            [spec.media_type for spec in self.digest_specs if spec.media_type]
        )

    def digest_spec_for_type(self, media_type):
        """Get the digest spec for a specific media type, if available.

        Parameters:
            media_type (str)
                a MIME type string identifying a type of image manifest or
                manifest list, e.g. "application/vnd.docker.distribution.manifest.v2+json"

        Returns:
            :class:`ContainerImageDigestPullSpec`
                if a digest of the requested type is available
            ``None``
                if a digest of the requested type is not available
        """
        for spec in self.digest_specs:
            if spec.media_type == media_type:
                return spec


@attr.s()
class ContainerImagePushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing a container image.

    This item represents any kind of image which can be pulled
    from a container image registry (i.e. an object identified by a
    ``NAME[:TAG|@DIGEST]`` string).

    In the case of a multi-arch image, one ContainerImagePushItem represents
    an image for a single architecture only.
    """

    dest_signing_key = attr.ib(
        type=str, default=None, converter=lambda s: s.lower() if s else None
    )
    """Desired signing key for this container image.

    Typically a 16-character identifier for a signing key, such
    as ``"199e2f91fd431d51"``, though the format is not enforced
    by this library.

    A non-empty value for this attribute should be interpreted as a request to
    use the specified key to generate signatures for this image's manifest(s).
    See `container-signature`_ docs for information on signatures.

    .. _container-signature: https://github.com/containers/image/blob/33bcba75bb181318608f989e18e086f0d83d254c/docs/containers-signature.5.md
    """

    source_tags = attr.ib(
        type=list, default=attr.Factory(frozenlist), converter=frozenlist
    )
    """Tags placed onto this image when it was built, if known.

    :type: List[str]
    """

    labels = attr.ib(type=dict, default=attr.Factory(frozendict), converter=frozendict)
    """Labels of this image, if known.

    This field is not guaranteed to include all labels associated with the image.

    :type: Dict[str, str]
    """

    arch = attr.ib(type=str, default=None)
    """Architecture of this image, if known.

    This field uses the conventional architecture strings used throughout the
    container ecosystem, such as ``"amd64"``.
    """

    pull_info = attr.ib(
        type=ContainerImagePullInfo,
        default=None,
        validator=instance_of(ContainerImagePullInfo),
    )
    """Metadata for pulling this image from a registry."""


@attr.s()
class SourceContainerImagePushItem(ContainerImagePushItem):
    """A :class:`~pushsource.PushItem` representing a source container image.

    Source container images are a special type of image which are not runnable
    but instead contain the packaged source code of a related binary image.

    See `this article <https://access.redhat.com/articles/3410171>`_ for more
    information on source container images.
    """


@attr.s()
class OperatorManifestPushItem(PushItem):
    """A :class:`~pushsource.PushItem` representing an operator manifests archive
    (typically named ``operator_manifests.zip``).
    """



import inspect
import functools

from six.moves.urllib import parse


def getfullargspec(x):
    # Helper for py2 vs py3 differences.

    if not hasattr(inspect, "getfullargspec"):  # pragma: no cover
        # 1. Must use older, deprecated getargspec.
        # 2. getfullargspec works fine when called on a class and
        #    returns spec of constructor, but getargspec complains.
        #    Make it work more like getfullargspec.
        if isinstance(x, type):
            x = x.__init__
        return inspect.getargspec(x)  # pylint: disable=deprecated-method

    return inspect.getfullargspec(x)


class SourceUrlError(ValueError):
    """Errors of this type are raised when an invalid URL is provided
    to :meth:`~pushsource.Source.get` and related methods.
    """


class SourceWrapper(object):
    # Internal class to ensure that all source instances support enter/exit
    # for with statements even if underlying instance doesn't implement it
    # (since it originally was not mandatory).
    def __init__(self, delegate):
        self.__delegate = delegate

    def __iter__(self):
        return self.__delegate.__iter__()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @classmethod
    def _maybe_wrap(cls, delegate):
        if hasattr(delegate, "__enter__"):
            # no wrapping needed
            return delegate
        # wrap with a no-op enter/exit
        return cls(delegate)


class Source(object):
    """A source of push items.

    This base class defines the interface for all pushsource backends.
    Instances of a specific backend can be obtained using
    the :meth:`~pushsource.Source.get` method.

    Though not mandatory, instances of ``Source`` are preferably used
    via ``with`` statements to ensure that all resources are cleaned up
    when no longer needed, as in example:

    .. code-block:: python

        with Source.get('some-url') as source:
            for item in source:
                do_something(item)

    Note that the items produced by a source are not bound to the lifecycle
    of a source instance, so it's safe to store them and continue using
    them beyond the end of the ``with`` statement.
    """

    _BACKENDS = {}
    _BACKENDS_BUILTIN = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __iter__(self):
        """Iterate over the push items contained within this source.

        Yields a series of :class:`~pushsource.PushItem` instances.
        """
        raise NotImplementedError()

    @classmethod
    def get(cls, source_url, **kwargs):
        """Obtain a push source from the given URL.

        Parameters:
            source_url (str)
                Specifies a push source backend and associated arguments.
                For information about push source URLs, see
                :ref:`urls`.

            kwargs (dict)
                Any additional keyword arguments to be passed into
                the backend.

        Raises:
            SourceUrlError
                If ``source_url`` is not a valid push source URL.

        Returns:
            :class:`~pushsource.Source`
                A new Source instance initialized with the given arguments.
        """
        return cls.get_partial(source_url, **kwargs)()

    @classmethod
    def get_partial(cls, source_url, **kwargs):
        """Obtain a push source constructor from the given URL and arguments.

        This method returns a constructor for a push source (rather than a push source
        instance directly) so that additional arguments may be provided later.

        One of the primary uses of this method is to provide your own preconfigured
        aliases for existing backends, as described in :ref:`binding`.

        Parameters:
            source_url (str)
                Specifies a push source backend and associated arguments.
                For information about push source URLs, see
                :ref:`urls`.

            kwargs (dict)
                Any additional keyword arguments to be passed into
                the backend.

        Raises:
            SourceUrlError
                If ``source_url`` is not a valid push source URL.

        Returns:
            callable
                A callable which accepts any number of keyword arguments
                and returns a :class:`~pushsource.Source`.
        """
        parsed = parse.urlparse(source_url)
        scheme = parsed.scheme

        if not scheme:
            raise SourceUrlError("Not a valid source URL: %s" % source_url)

        if scheme not in cls._BACKENDS:
            raise SourceUrlError(
                "Requested source '%s' but there is no registered backend '%s'"
                % (source_url, scheme)
            )

        klass = cls._BACKENDS[scheme]

        query = parsed.query

        # To make simple URLs less ugly, we'll allow the '?' to be omitted if needed.
        # Example: if "fedkoji" is a source pointing at fedora koji,
        # allow the caller to pass: fedkoji:rpm=python3-3.7.5-2.fc31.x86_64.rpm
        # instead of requiring:     fedkoji:?rpm=python3-3.7.5-2.fc31.x86_64.rpm
        if not query and not parsed.netloc and "=" in parsed.path:
            query = parsed.path

        url_kwargs = parse.parse_qs(query)

        # parse_qs forces everything into lists even if only a single term was given.
        #
        # e.g. arg1=foo&arg1=bar&arg2=baz => {'arg1': ['foo', 'bar'], 'arg2': ['baz']}
        #
        # This is annoying to deal with in source backends, so we'll unwrap anything
        # which had only a single element.
        for key in url_kwargs.keys():
            value = url_kwargs[key]
            if isinstance(value, list) and len(value) == 1:
                url_kwargs[key] = value[0]

        sig = getfullargspec(klass)

        if "url" in sig.args and parsed.path is not query:
            # If the source accepts a url argument, then the 'path' part
            # of the URL we were provided is itself required to be a URL.
            #
            # For example:
            #
            #  errata:https://errata.example.com/foo/bar?arg1=val&arg2=val...
            #
            # The "https://errata.example.com/foo/bar" part was parsed as 'path'
            # above, and is now provided to the source as a URL.
            url_kwargs["url"] = parsed.path

        # Coerce some standard arguments to the right type.
        for (key, converter) in [("threads", int), ("timeout", int)]:
            if key in url_kwargs:
                url_kwargs[key] = converter(url_kwargs[key])

        # TODO: check for some common mistakes here?

        url_kwargs.update(kwargs)

        @functools.wraps(klass)
        def partial_source(*inner_args, **inner_kwargs):
            kwargs = url_kwargs.copy()
            kwargs.update(inner_kwargs)
            return SourceWrapper._maybe_wrap(klass(*inner_args, **kwargs))

        return partial_source

    @classmethod
    def register_backend(cls, name, factory):
        """Register a new pushsource backend.

        This method allows registering additional backends beyond those
        shipped with the pushsource library. See :ref:`implementing` for
        more information.

        Parameters:
            name (str)
                The name of a backend. This should be a brief unique identifying
                string.

                If a backend of the given name is already registered, it will be
                overwritten.

            factory (callable)
                A callable used to create new instances of the backend.
                When invoked, this callable must return an object which implements
                the :class:`~pushsource.Source` interface.

        Raises:
            TypeError
                If ``factory`` is not callable.
        """
        if not callable(factory):
            raise TypeError("expected callable, got: %s" % repr(factory))

        cls._BACKENDS[name] = factory

    @classmethod
    def _register_backend_builtin(cls, name, factory):
        # Private equivalent of register_backend which also flags the
        # backend as built-in, i.e. it will be restored on a call to reset.
        cls.register_backend(name, factory)
        cls._BACKENDS_BUILTIN[name] = factory

    @classmethod
    def reset(cls):
        """Reset the library to the default configuration.

        This method will undo the effect of any prior calls to
        :meth:`~pushsource.Source.register_backend`, restoring only the
        default backends provided by the pushsource library.

        This may be used from within tests to ensure a known state.

        .. versionadded:: 2.6.0
        """
        cls._BACKENDS = cls._BACKENDS_BUILTIN.copy()



import os
import logging

from ...model import AmiRelease, AmiPushItem, AmiBillingCodes
from .staged_base import StagedBaseMixin, handles_type

LOG = logging.getLogger("pushsource")


class StagedAmiMixin(StagedBaseMixin):
    @handles_type("AWS_IMAGES")
    def __push_item(self, leafdir, metadata, entry):
        relative_path = os.path.join(leafdir.dest, leafdir.file_type, entry.name)
        file_md = metadata.file_metadata_or_die(relative_path)
        attributes = file_md.attributes
        release_md = attributes["release"]

        release_attrs = [
            "product",
            "date",
            "arch",
            "respin",
            "version",
            "base_product",
            "base_version",
            "variant",
            "type",
        ]
        release_kwargs = {}
        for key in release_attrs:
            if key in release_md:
                release_kwargs[key] = release_md[key]

        release = AmiRelease(**release_kwargs)

        image_kwargs = {"release": release}

        billing_codes_md = attributes.get("billing_codes") or {}
        if billing_codes_md:
            billing_codes_attrs = ["name", "codes"]
            billing_codes_kwargs = {}

            for key in billing_codes_attrs:
                if key in billing_codes_md:
                    billing_codes_kwargs[key] = billing_codes_md[key]

            billing_codes = AmiBillingCodes(**billing_codes_kwargs)
            image_kwargs.update({"billing_codes": billing_codes})

        image_attrs = [
            "type",
            "region",
            "virtualization",
            "volume",
            "root_device",
            "description",
            "sriov_net_support",
            "ena_support",
        ]

        for key in image_attrs:
            if key in attributes:
                image_kwargs[key] = attributes[key]

        return AmiPushItem(
            name=entry.name,
            src=entry.path,
            origin=leafdir.topdir,
            dest=[leafdir.dest],
            **image_kwargs
        )



import logging
import threading
from functools import partial

from six.moves import xmlrpc_client
from more_executors import Executors
from more_executors.futures import f_zip, f_map

from ...compat_attr import attr

LOG = logging.getLogger("pushsource.errata_client")


@attr.s()
class ErrataRaw(object):
    # Helper to collect raw responses of all ET APIs for a single advisory
    advisory_cdn_metadata = attr.ib(type=dict)
    advisory_cdn_file_list = attr.ib(type=dict)
    advisory_cdn_docker_file_list = attr.ib(type=dict)
    ftp_paths = attr.ib(type=dict)


class ErrataClient(object):
    def __init__(self, threads, url, **retry_args):
        self._executor = (
            Executors.thread_pool(name="pushsource-errata-client", max_workers=threads)
            .with_retry(**retry_args)
            .with_cancel_on_shutdown()
        )
        self._url = url
        self._tls = threading.local()

        self._get_advisory_cdn_metadata = partial(
            self._call_et, "get_advisory_cdn_metadata"
        )
        self._get_advisory_cdn_file_list = partial(
            self._call_et, "get_advisory_cdn_file_list"
        )
        self._get_advisory_cdn_docker_file_list = partial(
            self._call_et, "get_advisory_cdn_docker_file_list"
        )
        self._get_ftp_paths = partial(self._call_et, "get_ftp_paths")

    def shutdown(self):
        self._executor.shutdown(True)

    @property
    def _errata_service(self):
        # XML-RPC client connected to errata_service.
        # Each thread uses a separate client.
        if not hasattr(self._tls, "errata_service"):
            LOG.debug("Creating XML-RPC client for Errata Tool: %s", self._url)
            self._tls.errata_service = xmlrpc_client.ServerProxy(self._url)
        return self._tls.errata_service

    def _log_queried_et(self, response, advisory_id):
        # This message is visible by default for all advisories.
        # For more detailed logs of each individual call, see DEBUG logs
        # in _call_et.
        LOG.info("Queried Errata Tool for %s", advisory_id)
        return response

    def get_raw_f(self, advisory_id):
        """Returns Future[ErrataRaw] holding all ET responses for a particular advisory."""
        all_responses = f_zip(
            self._executor.submit(self._get_advisory_cdn_metadata, advisory_id),
            self._executor.submit(self._get_advisory_cdn_file_list, advisory_id),
            self._executor.submit(self._get_advisory_cdn_docker_file_list, advisory_id),
            self._executor.submit(self._get_ftp_paths, advisory_id),
        )
        all_responses = f_map(
            all_responses, partial(self._log_queried_et, advisory_id=advisory_id)
        )
        return f_map(all_responses, lambda tup: ErrataRaw(*tup))

    def _call_et(self, method_name, advisory_id):
        method = getattr(self._errata_service, method_name)

        # These APIs have had performance issues occasionally, so let's set up some
        # detailed structured logs which can be used to check the performance.
        LOG.debug(
            "Calling Errata Tool %s(%s)",
            method_name,
            advisory_id,
            extra={
                "event": {
                    "type": "errata-tool-call-start",
                    "method": method_name,
                    "advisory": advisory_id,
                }
            },
        )
        try:
            out = method(advisory_id)
            LOG.debug(
                "Errata Tool completed call %s(%s)",
                method_name,
                advisory_id,
                extra={
                    "event": {
                        "type": "errata-tool-call-end",
                        "method": method_name,
                        "advisory": advisory_id,
                    }
                },
            )
            return out
        except:
            LOG.debug(
                "Failed to call Errata Tool %s(%s)",
                method_name,
                advisory_id,
                extra={
                    "event": {
                        "type": "errata-tool-call-fail",
                        "method": method_name,
                        "advisory": advisory_id,
                    }
                },
            )
            raise



import logging
import os

from ...model import FilePushItem
from .staged_base import StagedBaseMixin, handles_type

LOG = logging.getLogger("pushsource")


class StagedFilesMixin(StagedBaseMixin):
    @handles_type("ISOS")
    @handles_type("FILES")
    def __file_push_item(self, leafdir, metadata, entry):
        relative_path = os.path.join(leafdir.dest, leafdir.file_type, entry.name)
        file_md = metadata.file_metadata_or_die(relative_path)

        return FilePushItem(
            name=file_md.filename or entry.name,
            src=entry.path,
            description=file_md.attributes.get("description"),
            version=file_md.version,
            sha256sum=file_md.sha256sum,
            origin=leafdir.topdir,
            dest=[leafdir.dest],
        )



import logging

import jsonschema

from .schema import get_schema


LOG = logging.getLogger("pushsource")


class Validator(object):
    """A helper to validate jsonschemas with some friendlier
    log messages applied on validation errors.
    """

    def __init__(self, schema_name, ids=None):
        """Obtain a validator.

        Arguments:

            schema_name (str)
                Name of schema within pushsource library to be used by
                this validator (e.g. "errata")

            ids (list[str])
                A list of attributes which can be used to identify user-meaningful
                objects in validated data.

                When validation fails, the innermost found object having any of these
                attributes will be considered the most relevant object to report to
                the user, and the attribute value will be included in a log message.

                Example: if validation fails on an object like this:

                  {'object1': {'filename': 'foobar', 'metadata': {... FAILS HERE}}}

                ... and if 'ids' contains 'filename', then the validation failure
                will be logged as "foobar is not valid".
        """
        self._schema = get_schema(schema_name)
        self._ids = ids or []

    def _get_subobject_label(self, data, error):
        # Given data which has failed validation plus the validation error,
        # try to obtain and return the most user-meaningful label for the
        # piece which failed validation.

        # This is where the error occurred, for example if:
        #      data = {'foo': [{'bar': <something invalid>}]}
        # then path = ['foo', 0, 'bar']
        #
        # note: 'absolute_path' was 'path' in jsonschema <= 2.3.0.
        path = list(getattr(error, "absolute_path", error.path))

        # Build up a list of objects we were validating.
        sub = data
        objects = [sub]
        while path:
            sub = sub[path[0]]
            objects.append(sub)
            path = path[1:]

        # Reverse the list, then we have the objects in order from
        # innermost to outermost being validated, e.g. if:
        #      data = {'foo': [{'bar': <something invalid>}]}
        # then
        #      objects = [<bar dict>, <list containing bar dict>, <foo dict>]
        objects.reverse()

        # Only dicts are able to be checked
        objects = [o for o in objects if isinstance(o, dict)]

        # The innermost object which contains one of our configured 'ids'
        # is the most meaningful object to the user.
        for o in objects:
            for label in self._ids:
                if o.get(label):
                    return o[label]

    def validate(self, data, data_label):
        """Validate an object against the configured schema.

        This method behaves the same as `jsonschema.validate`, except that
        if a validation error occurs, a user-oriented message is logged at
        ERROR level before an exception is raised.

        The idea is that a typical end-user has a better chance of understanding
        a basic error message pointing at which piece of data failed validation,
        than they do of understanding the jsonschema ValidationError in raw form.

        Arguments:

            data (object)
                Any object to be validated.

            data_label (str)
                A label for the data being validated.

                This should be a brief string which would make sense to the end-user
                which will help identify the data failing validation.

                For example, if you are validating JSON loaded from a file,
                passing the filename as data_label would be reasonable - it will tell
                the user which file failed validation.
        """

        try:
            jsonschema.validate(data, self._schema)
        except jsonschema.ValidationError as error:
            label = self._get_subobject_label(data, error)

            if label:
                # We found some meaningful label for the particular
                # item failing validation - great, mention that object
                # (and also the containing object)
                label = "%s (in %s)" % (label, data_label)
            else:
                # We didn't find anything, oh well...
                # Just mention the outermost object.
                label = data_label

            LOG.error("%s is not valid", label)

            raise



# Helpers for py2 vs py3 compatibility issues.

import sys

try:
    from os import scandir
except ImportError:  # pragma: no cover
    # expected path for python <3.5
    from scandir import scandir

if sys.version_info >= (3, 0):
    # It is preferred to use immutable dicts when possible
    from frozendict import frozendict
else:  # pragma: no cover
    # But it's not available on py2, fall back to plain old dicts there
    frozendict = dict


__all__ = ["scandir", "frozendict"]



import os
import logging
from concurrent import futures

import six
from six.moves.urllib import parse
from more_executors import Executors

from .errata_client import ErrataClient

from ... import compat_attr as attr
from ...source import Source
from ...model import (
    ErratumPushItem,
    ContainerImagePushItem,
    ModuleMdSourcePushItem,
    RpmPushItem,
    OperatorManifestPushItem,
    conv,
)
from ...helpers import (
    list_argument,
    try_bool,
    force_https,
    as_completed_with_timeout_reset,
)

LOG = logging.getLogger("pushsource")


class ErrataSource(Source):
    """Uses an advisory from Errata Tool as the source of push items."""

    def __init__(
        self,
        url,
        errata,
        koji_source=None,
        rpm_filter_arch=None,
        legacy_container_repos=False,
        threads=4,
        timeout=60 * 60 * 4,
    ):
        """Create a new source.

        Parameters:
            url (src)
                Base URL of Errata Tool, e.g. "http://errata.example.com",
                "https://errata.example.com:8123".

            errata (str, list[str])
                Advisory ID(s) to be used as push item source.
                If a single string is given, multiple IDs may be
                comma-separated.

            koji_source (str)
                URL of a koji source associated with this Errata Tool
                instance.

            rpm_filter_arch (str, list[str])
                If provided, only RPMs for these given arch(es) will be produced;
                e.g. "x86_64", "src" and "noarch".

            legacy_container_repos (bool)
                If ``True``, any container push items generated by this source will
                use a legacy format for repository IDs.

                This is intended to better support certain legacy code and will be
                removed when no longer needed. Only use this if you know that you
                need it.

            threads (int)
                Number of threads used for concurrent queries to Errata Tool
                and koji.

            timeout (int)
                Number of seconds after which an error is raised, if no progress is
                made during queries to Errata Tool.
        """
        self._url = force_https(url)
        self._errata = list_argument(errata)
        self._client = ErrataClient(threads=threads, url=self._errata_service_url)

        self._rpm_filter_arch = list_argument(rpm_filter_arch, retain_none=True)

        # This executor doesn't use retry because koji & ET executors already do that.
        self._executor = Executors.thread_pool(
            name="pushsource-errata", max_workers=threads
        ).with_cancel_on_shutdown()

        # We set aside a separate thread pool for koji so that there are separate
        # queues for ET and koji calls, yet we avoid creating a new thread pool for
        # each koji source.
        self._koji_executor = (
            Executors.thread_pool(name="pushsource-errata-koji", max_workers=threads)
            .with_retry()
            .with_cancel_on_shutdown()
        )
        self._koji_cache = {}
        self._koji_source_url = koji_source

        self._legacy_container_repos = try_bool(legacy_container_repos)
        self._timeout = timeout

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._client.shutdown()
        self._executor.shutdown(True)
        self._koji_executor.shutdown(True)

    @property
    def _errata_service_url(self):
        # URL for the errata_service XML-RPC endpoint provided by ET.
        #
        # Note the odd handling of scheme here. The reason is that
        # ET oddly provides different APIs over http and https.
        #
        # This XML-RPC API is available anonymously over *https* only,
        # but for compatibility the caller is allowed to provide http
        # or https scheme and we'll apply the scheme we know is actually
        # needed.
        #
        # The XML-RPC endpoint is no longer available over http as of
        # 11 Oct 2021. This class's constructor will still accept either
        # http or https, but will force the given url to use https going
        # forward.
        parsed = parse.urlparse(self._url)
        base = "{0.scheme}://{0.netloc}{0.path}".format(parsed)

        # Note: os.path.join will join these components with a '\' on Windows
        # systems, which is not intended. The tradeoff is that os.path.join
        # handles cases where path components end with trailing path separators
        # and doesn't doesn't duplicate them i.e. prevents something like
        # https://errata.example.com/some/path//errata/errata_service
        return os.path.join(base, "errata/errata_service")

    def _koji_source(self, **kwargs):
        if not self._koji_source_url:
            raise ValueError("A Koji source is required but none is specified")
        return Source.get(
            self._koji_source_url,
            cache=self._koji_cache,
            executor=self._koji_executor,
            **kwargs
        )

    @property
    def _advisory_ids(self):
        # TODO: other cases (comma-separated; plain string)
        return self._errata

    def _push_items_from_raw(self, raw):
        erratum = ErratumPushItem._from_data(raw.advisory_cdn_metadata)

        items = self._push_items_from_rpms(erratum, raw.advisory_cdn_file_list)

        # The erratum should go to all the same destinations as the rpms,
        # before FTP paths are added.
        erratum_dest = set(erratum.dest or [])
        for item in items:
            for dest in item.dest:
                erratum_dest.add(dest)
        erratum = attr.evolve(erratum, dest=sorted(erratum_dest))

        # Adjust push item destinations according to FTP paths from ET, if any.
        items = self._add_ftp_paths(items, erratum, raw.ftp_paths)

        items = items + self._push_items_from_container_manifests(
            erratum, raw.advisory_cdn_docker_file_list
        )

        return [erratum] + items

    def _push_items_from_container_manifests(self, erratum, docker_file_list):
        if not docker_file_list:
            return []

        # Example of container list for one item to one repo:
        #
        # {
        #     "dotnet-21-container-2.1-77.1621419388": {
        #         "docker": {
        #             "target": {
        #                 "external_repos": {
        #                     "rhel8/dotnet-21": {
        #                         "container_full_sig_key": "199e2f91fd431d51",
        #                         "container_sig_key": "fd431d51",
        #                         "tags": ["2.1", "2.1-77.1621419388", "latest"],
        #                     },
        #                 },
        #                 "repos": /* like external_repos but uses pulp repo IDs */,
        #             }
        #         }
        #     }
        # }
        #

        # We'll be getting container metadata from these builds.
        with self._koji_source(
            container_build=list(docker_file_list.keys())
        ) as koji_source:

            out = []

            for item in koji_source:
                if isinstance(item, ContainerImagePushItem):
                    item = self._enrich_container_push_item(
                        erratum, docker_file_list, item
                    )
                elif isinstance(item, OperatorManifestPushItem):
                    # Accept this item but nothing special to do
                    pass
                else:
                    # If build contained anything else, ignore it
                    LOG.debug(
                        "Erratum %s: ignored unexpected item from koji source: %s",
                        erratum.name,
                        item,
                    )
                    continue

                item = attr.evolve(item, origin=erratum.name)
                out.append(item)

        return out

    def _enrich_container_push_item(self, erratum, docker_file_list, item):
        # metadata from koji doesn't contain info about where the image should be
        # pushed and a few other things - enrich it now
        errata_meta = docker_file_list.get(item.build) or {}
        target = (errata_meta.get("docker") or {}).get("target") or {}

        repos = target.get("external_repos") or {}

        if self._legacy_container_repos:
            repos = target.get("repos") or {}

        sig_keys = set()
        dest = []
        for repo_id, repo_data in repos.items():
            for tag in repo_data.get("tags") or []:
                dest.append("%s:%s" % (repo_id, tag))

            sig_key = repo_data.get("container_full_sig_key")
            if sig_key:
                sig_keys.add(sig_key)

        dest = sorted(set(dest))

        # If ET is not requesting to push this to any repos or tags at all,
        # it's considered an error.
        if not dest:
            raise ValueError(
                "Erratum %s requests container build %s but provides no repositories"
                % (erratum.name, item.build)
            )

        if len(sig_keys) > 1:
            # The API structure in Errata Tool would theoretically allow for multiple
            # signing keys on a single image. However, it's unclear if there's any use-case
            # for it, and all existing code in Pub for dealing with container images does not
            # even come close to handling it correctly, so we're going to flag this as
            # unsupported for now.
            raise ValueError(
                "Unsupported: erratum %s requests multiple signing keys (%s) on build %s"
                % (erratum.name, ", ".join(sorted(sig_keys)), item.build)
            )

        dest_signing_key = None if not sig_keys else list(sig_keys)[0]

        # koji source provided basic info on container image, ET provides policy on
        # where/how it should be pushed, combine them both to get final push item
        return attr.evolve(item, dest=dest, dest_signing_key=dest_signing_key)

    def _push_items_from_rpms(self, erratum, rpm_list):
        out = []

        for build_nvr, build_info in six.iteritems(rpm_list):
            out.extend(self._rpm_push_items_from_build(erratum, build_nvr, build_info))
            out.extend(
                self._module_push_items_from_build(erratum, build_nvr, build_info)
            )

        return out

    def _module_push_items_from_build(self, erratum, build_nvr, build_info):
        modules = (build_info.get("modules") or {}).copy()

        module_filenames = list(modules.keys())

        # We always request the modulemd.src.txt because we might need it later
        # depending on the ftp_paths response.
        module_filenames.append("modulemd.src.txt")

        out = []

        # Get a koji source which will yield all modules from the build
        with self._koji_source(
            module_build=[build_nvr], module_filter_filename=module_filenames
        ) as koji_source:

            for push_item in koji_source:
                # ET uses filenames to identify the modules here, we must do the same.
                basename = os.path.basename(push_item.src)
                dest = modules.pop(basename, [])

                # Fill in more push item details based on the info provided by ET.
                push_item = attr.evolve(push_item, dest=dest, origin=erratum.name)

                out.append(push_item)

        # Were there any requested modules we couldn't find?
        missing_modules = ", ".join(sorted(modules.keys()))
        if missing_modules:
            msg = "koji build {nvr} does not contain {missing} (requested by advisory {erratum})".format(
                nvr=build_nvr, missing=missing_modules, erratum=erratum.name
            )
            raise ValueError(msg)

        return out

    def _filter_rpms_by_arch(self, erratum, rpm_filenames):
        if self._rpm_filter_arch is None:
            return rpm_filenames

        out = []
        ok_arches = [conv.archstr(arch) for arch in self._rpm_filter_arch]

        for filename in rpm_filenames:
            components = filename.split(".")
            if len(components) >= 3 and components[-1] == "rpm":
                arch = components[-2]
                if conv.archstr(arch) in ok_arches:
                    out.append(filename)
                    continue

            LOG.debug(
                "Erratum %s: RPM removed by arch filter: %s", erratum.name, filename
            )

        return out

    def _rpm_push_items_from_build(self, erratum, build_nvr, build_info):
        rpms = build_info.get("rpms") or {}
        signing_key = build_info.get("sig_key") or None
        sha256sums = (build_info.get("checksums") or {}).get("sha256") or {}
        md5sums = (build_info.get("checksums") or {}).get("md5") or {}

        rpm_filenames = self._filter_rpms_by_arch(erratum, list(rpms.keys()))

        # Get a koji source which will yield all desired push items from this build.
        koji_source = self._koji_source(rpm=rpm_filenames, signing_key=signing_key)

        out = []

        for push_item in koji_source:
            # Do not allow to proceed if RPM was absent
            if push_item.state == "NOTFOUND":
                raise ValueError(
                    "Advisory refers to %s but RPM was not found in koji"
                    % push_item.name
                )

            # Note, we can't sanity check here that the push item's build
            # equals ET's NVR, because it's not always the case.
            #
            # Example:
            #  RPM: pgaudit-debuginfo-1.4.0-4.module+el8.1.1+4794+c82b6e09.x86_64.rpm
            #  belongs to build: 1015162 (pgaudit-1.4.0-4.module+el8.1.1+4794+c82b6e09)
            #  but ET refers instead to module build: postgresql-12-8010120191120141335.e4e244f9.
            #
            # We also make use of this to fill in the module_build attribute on items when
            # available.
            #
            # (This is not ideal because we don't really "know" that a non-matching NVR here is
            # the module build NVR, we are relying on the ET implementation detail that this is
            # the only reason they should not match; though legacy code already depended on this
            # for years, so maybe it's fine. We also have the heuristic of scanning for 'module'
            # in the NVR to make exceptions less likely.)
            module_build = None
            if push_item.build != build_nvr and ".module" in push_item.build:
                module_build = build_nvr

            # Fill in more push item details based on the info provided by ET.
            push_item = attr.evolve(
                push_item,
                sha256sum=sha256sums.get(push_item.name),
                md5sum=md5sums.get(push_item.name),
                dest=rpms.get(push_item.name),
                origin=erratum.name,
                module_build=module_build,
            )

            out.append(push_item)

        return out

    def _add_ftp_paths(self, items, erratum, ftp_paths):
        # ftp_paths structure is like this:
        #
        # {
        #     "xorg-x11-server-1.20.4-16.el7_9": {
        #         "rpms": {
        #             "xorg-x11-server-1.20.4-16.el7_9.src.rpm": [
        #                 "/ftp/pub/redhat/linux/enterprise/7Client/en/os/SRPMS/",
        #                 "/ftp/pub/redhat/linux/enterprise/7ComputeNode/en/os/SRPMS/",
        #                 "/ftp/pub/redhat/linux/enterprise/7Server/en/os/SRPMS/",
        #                 "/ftp/pub/redhat/linux/enterprise/7Workstation/en/os/SRPMS/",
        #             ]
        #         },
        #         "modules": [
        #            "/ftp/pub/redhat/linux/enterprise/AppStream-8.0.0.Z/en/os/modules/"
        #         ],
        #         "sig_key": "fd431d51",
        #     }
        # }
        #
        # We use the (rpm, module) => ftp path mappings, which should be added onto
        # our existing push items if they match.
        #
        rpm_to_paths = {}
        build_to_module_paths = {}
        builds_need_modules = set()
        builds_have_modules = set()
        for build_nvr, build_map in ftp_paths.items():
            for (rpm_name, paths) in (build_map.get("rpms") or {}).items():
                rpm_to_paths[rpm_name] = paths

            modules = build_map.get("modules") or []
            build_to_module_paths[build_nvr] = modules
            if modules:
                builds_need_modules.add(build_nvr)

        out = []
        for item in items:
            if isinstance(item, RpmPushItem):
                # RPMs have dest updated with FTP paths.
                paths = rpm_to_paths.get(item.name) or []
                item = attr.evolve(item, dest=item.dest + paths)
                out.append(item)
            elif isinstance(item, ModuleMdSourcePushItem):
                # modulemd sources have dest updated with FTP paths and will
                # also be filtered out if there are no matches at all (because
                # modulemd sources aren't delivered in any other manner)
                paths = build_to_module_paths.get(item.build)
                if paths:
                    builds_have_modules.add(item.build)
                    item = attr.evolve(item, dest=item.dest + paths)
                if item.dest:
                    out.append(item)
                else:
                    # modulemd sources are filtered out altogether if ET did
                    # not provide any destinations.
                    LOG.debug(
                        "Erratum %s: modulemd source skipped due to no destinations: %s",
                        erratum.name,
                        item.src,
                    )
            else:
                # Other types of items are unaffected by ftp_paths.
                out.append(item)

        # If ET requests that modules should be pushed for any builds and we're
        # missing a modulemd.src.txt for those, it's a fatal error.
        builds_missing_modules = sorted(builds_need_modules - builds_have_modules)
        if builds_missing_modules:
            msg = "Erratum %s: missing modulemd sources on koji build(s): %s" % (
                erratum.name,
                ", ".join(builds_missing_modules),
            )
            raise ValueError(msg)

        return out

    def __iter__(self):
        # Get raw ET responses for all errata.
        raw_fs = [self._client.get_raw_f(id) for id in self._advisory_ids]

        # Convert them to lists of push items
        push_items_fs = []
        for f in futures.as_completed(raw_fs, timeout=self._timeout):
            push_items_fs.append(
                self._executor.submit(self._push_items_from_raw, f.result())
            )

        completed_fs = as_completed_with_timeout_reset(
            push_items_fs, timeout=self._timeout
        )
        for f in completed_fs:
            for pushitem in f.result():
                yield pushitem


Source._register_backend_builtin("errata", ErrataSource)



import base64
import hashlib
import json

import os
import six

try:
    from json.decoder import JSONDecodeError

    JSONException = JSONDecodeError
except ImportError:  # pragma: no cover
    JSONException = ValueError

from six.moves.urllib import parse as urlparse
from requests.adapters import HTTPAdapter

# pylint: disable-next=import-error
from requests.packages.urllib3.util import Retry
from requests import Session, exceptions

try:
    import urllib2 as request
except ImportError:
    # Yes, there's a six.moves for this import, but,
    # parse_http_list and parse_keqv_list are no available
    # in python-six package from RHEL6, these alias
    # were included in a recent six version.
    from urllib import request


MEDIATYPE_SCHEMA2_V1 = "application/vnd.docker.distribution.manifest.v1+json"
MEDIATYPE_SCHEMA2_V1_SIGNED = (
    "application/vnd.docker.distribution.manifest.v1+prettyjws"
)
MEDIATYPE_SCHEMA2_V2 = "application/vnd.docker.distribution.manifest.v2+json"
MEDIATYPE_SCHEMA2_V2_LIST = "application/vnd.docker.distribution.manifest.list.v2+json"


class AuthToken(object):
    def __init__(self, token=None):
        self.token = token


def update_auth_header(headers, token):
    """
    Adds the token into the request's headers as specified in the Docker v2 API documentation.
    https://docs.docker.com/registry/spec/auth/token/#using-the-bearer-token
    """
    headers.update({"authorization": "Bearer %s" % token})
    return headers


def request_token(session, response, credentials):
    """
    Attempts to retrieve the correct token based on the 401 response header.
    According to the Docker API v2 documentation, the token be retrieved by issuing a GET
    request to the url specified by the `realm` within the `WWW-Authenticate` header. This
    request should add the following query parameters:
        service: the name of the service that hosts the desired resource
        scope:   the specific resource and permissions requested
    https://docs.docker.com/registry/spec/auth/token/#requesting-a-token
    """
    auth_info = parse_401_response_headers(response.headers)
    try:
        token_url = auth_info.pop("realm")
    except KeyError as e:
        six.raise_from(IOError("No realm specified for token auth challenge."), e)

    parse_result = urlparse.urlparse(token_url)
    query_dict = urlparse.parse_qs(parse_result.query)
    query_dict.update(auth_info)
    url_pieces = list(parse_result)
    url_pieces[4] = urlparse.urlencode(query_dict)
    token_url = urlparse.urlunparse(url_pieces)

    http_opts = {}
    if credentials:
        # whenever token is requested, we need to send original
        # credentials again. On the other hand we don't want them stored
        # int http_opts, as it would override Bearer token
        http_opts["auth"] = credentials

    session.mount(
        token_url, HTTPAdapter(max_retries=Retry(total=3, status_forcelist=[500, 404]))
    )
    session.headers.pop("authorization", None)
    response = session.get(token_url, **http_opts)
    response.raise_for_status()

    return response.json()["token"]


def parse_401_response_headers(response_headers):
    """
    Parse the headers from a 401 response into a dictionary that contains the information
    necessary to retrieve a token.
    Example:
    Www-Authenticate: Bearer realm="https://auth.docker.io/token",
    service="registry.docker.io",scope="repository:library/nginx:pull,push"
    """
    auth_header = response_headers.get("www-authenticate")
    if auth_header is None:
        raise IOError(
            "401 responses are expected to contain authentication information"
        )
    auth_header = auth_header[len("Bearer ") :]

    # The remaining string consists of comma separated key=value pairs
    # according to RFC 2617
    try:
        items = request.parse_http_list(auth_header)
        return request.parse_keqv_list(items)
    except ValueError as e:
        six.raise_from(
            IOError("401 responses are expected to contain authentication information"),
            e,
        )


def registry_request(
    session,
    uri,
    action="get",
    auth_token=None,
    retry_404=True,
    headers=None,
    data=None,
    credentials=None,
    http_opts=None,
):
    """
    Retrieve a single path within the upstream registry, and return a
    2-tuple of the headers and the response body.
    """
    status_forcelist = [500]
    if retry_404:
        status_forcelist.append(404)
    if headers is None:
        headers = {}
    if http_opts is None:
        http_opts = {}
    assert isinstance(headers, dict)
    http_opts = http_opts or {}

    session.mount(
        uri,
        HTTPAdapter(
            max_retries=Retry(
                total=3,
                backoff_factor=1.0,  # 1.0, 2.0, 4.0, 8.0, ...
                status_forcelist=status_forcelist,
            )
        ),
    )
    if auth_token.token:
        update_auth_header(headers, auth_token.token)

    _request = getattr(session, action)
    try:
        response = _request(uri, headers=headers, data=data, **http_opts)
        response.raise_for_status()
    except exceptions.HTTPError as e:
        if e.response.status_code == 401:
            auth_header = e.response.headers.get("www-authenticate") or ""
            auth_header = auth_header.lower()
            if auth_header.startswith("bearer"):
                auth_token.token = request_token(session, response, credentials)
                update_auth_header(headers, auth_token.token)
            elif auth_header.startswith("basic"):
                http_opts["auth"] = credentials
            else:
                raise
            response = _request(uri, headers=headers, data=data, **http_opts)
            response.raise_for_status()
        else:
            raise
    return response


def get_basic_auth(host, home=None):
    # Look for docker config file in home directory for username and password
    home_dir = os.path.expanduser("~")
    conf_file = os.path.join(home or home_dir, ".docker/config.json")
    if os.path.isfile(conf_file):
        with open(conf_file) as f:
            config = json.load(f)
            auth = config.get("auths", {}).get(host, {}).get("auth")
            if auth:
                return tuple(base64.b64decode(auth).decode().split(":"))
    return None, None


# Copied from https://github.com/pulp/pulp_docker.git
def _calculate_digest(raw_manifest, manifest):
    _raw_manifest = raw_manifest
    if "signatures" in manifest:
        protected = manifest["signatures"][0]["protected"]
        paddings = {0: "", 2: "==", 3: "="}
        protected += paddings[len(protected) % 4]
        unprotected = json.loads(base64.b64decode(protected))
        signed_length = unprotected["formatLength"]
        signed_tail = base64.b64decode(
            unprotected["formatTail"] + paddings[len(unprotected["formatTail"]) % 4]
        )
        _raw_manifest = raw_manifest[:signed_length] + signed_tail
    digest = "{a}:{d}".format(a="sha256", d=hashlib.sha256(_raw_manifest).hexdigest())
    return digest


def get_manifest(registry, repo, digest, manifest_types=None, token=None):
    auth = get_basic_auth(registry.split("://")[1])
    token = token or AuthToken()

    if manifest_types is None:
        manifest_types = [MEDIATYPE_SCHEMA2_V1, MEDIATYPE_SCHEMA2_V1_SIGNED]
    headers = {"Accept": ",".join(manifest_types)}
    session = Session()
    try:
        resp = registry_request(
            session,
            "%s/v2/%s/manifests/%s" % (registry, repo, digest),
            action="get",
            auth_token=token,
            retry_404=True,
            credentials=auth,
            headers=headers,
        )
        digest = resp.headers.get("docker-content-digest", None)
        if not digest:
            digest = _calculate_digest(resp.content, resp.json())

        content_type = resp.headers.get("Content-Type", None)
    except exceptions.HTTPError as e:
        try:
            json_error = e.response.json()
        except (ValueError, JSONException):
            json_error = {}
        if e.response.status_code == 404 and (
            "MANIFEST_UNKNOWN" in [err["code"] for err in json_error.get("errors", [])]
            or "TAG_EXPIRED" in [err["code"] for err in json_error.get("errors", [])]
        ):
            six.raise_from(
                KeyError(
                    "Failed to get the manifest for image '%s' [%s]" % (repo, digest)
                ),
                e,
            )
            # otherwise probably true 404 or other error, let's reraise it
        raise e
    return content_type, digest, resp.json()


def get_blob(registry, repo, digest, token=None):
    auth = get_basic_auth(registry.split("://")[1])
    token = token or AuthToken()
    session = Session()
    resp = registry_request(
        session,
        "%s/v2/%s/blobs/%s" % (registry, repo, digest),
        action="get",
        auth_token=token,
        retry_404=True,
        credentials=auth,
    )
    resp.raise_for_status()
    return resp


def api_version_check(registry, token=None, credentials=None):
    """
    Make a call to the registry URL's /v2/ API call to determine if the registry supports API
    v2.
    """
    token = token or AuthToken()
    session = Session()
    if not credentials:
        auth = get_basic_auth(registry.split("://")[1])
    else:
        auth = credentials
    try:
        response = registry_request(
            session, "%s/v2/" % registry, credentials=auth, auth_token=token
        )
    except exceptions.HTTPError as e:
        if e.response.status_code == 404:
            return False
        raise e
    try:
        version = response.headers["Docker-Distribution-API-Version"]
        if version != "registry/2.0":
            return False
    except KeyError:
        # If the Docker-Distribution-API-Version header isn't present, we will assume that this
        # is a valid Docker 2.0 API server so that simple file-based webservers can serve as our
        # remote feed.
        pass
    return True


def inspect(registry, repo, digest, token=None):
    token = token or AuthToken()

    manifest_type, digest, manifest = get_manifest(
        registry, repo, digest, manifest_types=[MEDIATYPE_SCHEMA2_V2_LIST], token=token
    )
    if manifest_type == MEDIATYPE_SCHEMA2_V2:
        inspected = get_blob(registry, repo, manifest["config"]["digest"]).json()
    elif manifest_type == MEDIATYPE_SCHEMA2_V2_LIST:
        manifest_type, _digest, manifest = get_manifest(
            registry,
            repo,
            manifest["manifests"][0]["digest"],
            manifest_types=[MEDIATYPE_SCHEMA2_V2],
            token=token,
        )
        inspected = get_blob(registry, repo, manifest["config"]["digest"]).json()
    else:
        inspected = {"architecture": manifest["architecture"], "labels": {}}
    if manifest_type == MEDIATYPE_SCHEMA2_V2 and not inspected.get("config", {}).get(
        "Labels"
    ):
        inspected["source"] = True
    inspected["digest"] = digest
    return inspected



# Helpers for dealing with container metadata from koji.

from ..model import ContainerImagePullSpec, ContainerImagePullInfo
from .. import compat_attr as attr

MIME_TYPE_MANIFEST_LIST = "application/vnd.docker.distribution.manifest.list.v2+json"


class ContainerArchiveHelper(object):
    def __init__(self, build, archive):
        self._build = build
        self._archive = archive

    @property
    def build_image(self):
        extra = self._build.get("extra") or {}
        typeinfo = extra.get("typeinfo") or {}
        # It is preferred to use the metadata under 'typeinfo' if present
        return typeinfo.get("image") or extra.get("image") or {}

    @property
    def build_index(self):
        return self.build_image.get("index") or {}

    @property
    def archive_extra(self):
        return self._archive.get("extra") or {}

    @property
    def archive_docker(self):
        return self.archive_extra.get("docker") or {}

    @property
    def source_tags(self):
        return self.archive_docker.get("tags") or []

    @property
    def arch(self):
        config = self.archive_docker.get("config") or {}
        image = self.archive_extra.get("image") or {}

        out = config.get("architecture")

        if not out:
            # Ideally we got the arch from 'config' which will use the
            # docker-native arch strings. For ancient/weird builds we may
            # have had to fall back to 'image'.
            out = image.get("arch")

            # In this case, we get koji's native arch strings instead
            # which can be different. Convert them to the terms used in
            # the container world so we get consistent output.
            archmap = {"x86_64": "amd64", "aarch64": "arm64"}
            out = archmap.get(out, out)

        return out

    @property
    def labels(self):
        out = {}

        config = self.archive_docker.get("config") or {}
        # yes, config.config is the expected structure
        config = config.get("config") or {}
        raw_labels = config.get("Labels") or {}

        # We are going to be a bit selective with the labels we return.
        # The reason for this is just to keep a higher level of control
        # over the data we produce. Labels include all sorts of random
        # stuff and it'd be best not to use them in some cases and instead
        # require proper fields to be added onto our models.
        #
        # "com.redhat." is permitted because fields in that namespace are
        # known to be required for some operator images, e.g.
        # com.redhat.openshift.versions documented at
        # https://redhat-connect.gitbook.io/certified-operator-guide/ocp-deployment/operator-metadata/bundle-directory
        # which has its own non-trivial syntax we don't want to get into parsing.
        for (key, value) in raw_labels.items():
            if key.startswith("com.redhat."):
                out[key] = value

        return out

    @property
    def pull_info(self):
        ################## Image pull specs - from archive ##############################
        # Get image pull info from archive.
        archive_raw_specs = self.archive_docker.get("repositories") or []
        archive_digests = self.archive_docker.get("digests") or {}
        image_digest_specs = get_digest_specs(archive_raw_specs, archive_digests)
        image_tag_specs = get_tag_specs(archive_raw_specs)

        ################## List pull specs - from build #################################
        build_raw_specs = self.build_index.get("pull") or {}
        build_digests = self.build_index.get("digests") or {}
        list_digest_specs = get_digest_specs(build_raw_specs, build_digests)
        list_tag_specs = get_tag_specs(build_raw_specs)

        # Metadata does not explicitly tell us the media type per tag, but these tag
        # specs are known to be for manifest lists. We'll sanity check that this type
        # is present in the digests map at least.
        if MIME_TYPE_MANIFEST_LIST in build_digests:
            list_tag_specs = [
                attr.evolve(spec, media_types=[MIME_TYPE_MANIFEST_LIST])
                for spec in list_tag_specs
            ]

        ################### Put it all together #######################################
        # Note: ContainerimagePullInfo itself performs various operations on these specs
        # such as sorting and de-duplication.
        return ContainerImagePullInfo(
            tag_specs=list_tag_specs + image_tag_specs,
            digest_specs=list_digest_specs + image_digest_specs,
        )


def get_tag_specs(raw_specs):
    return [ContainerImagePullSpec._from_str(s) for s in raw_specs if "@" not in s]


def get_digest_specs(raw_specs, digests_map):
    out = []

    # Reverse the type => digest map
    digest_to_type = {}
    for (media_type, digest) in digests_map.items():
        digest_to_type[digest] = media_type

    for raw_spec in raw_specs:
        if "@" not in raw_spec:
            continue

        # raw_spec is a value such as:
        # registry-proxy.engineering.redhat.com/rh-osbs/openshift3-ose-logging-elasticsearch5@sha256:14f09792420665534c9d66ee3d055ff5b78d95e736c4a9fc8172cc01ecf39598
        spec = ContainerImagePullSpec._from_str(raw_spec)

        # If we know the specific type of this digest, add it.
        # (It's only ancient builds where we won't know.)
        spec = attr.evolve(spec, media_type=digest_to_type.get(spec.digest))

        out.append(spec)

    return out



import logging
import json
import yaml

from .staged_base import StagedBaseMixin, handles_type
from ... import compat_attr as attr
from ...model import ErratumPushItem
from ...validator import Validator

LOG = logging.getLogger("pushsource")
VALIDATOR = Validator("errata", ["filename", "id"])


class StagedErrataMixin(StagedBaseMixin):
    @handles_type("ERRATA")
    def __make_push_item(self, leafdir, _, entry):
        with open(entry.path, "rt") as fh:
            if entry.path.endswith(".json"):
                raw = json.load(fh)
            else:
                raw = yaml.safe_load(fh)

        # This field can be provided when data comes from ET,
        # but NOT when using staged files, as the staging structure
        # itself encodes the destinations.
        raw.pop("cdn_repo", None)

        VALIDATOR.validate(raw, entry.path)

        item = ErratumPushItem._from_data(raw)
        return attr.evolve(
            item, origin=leafdir.topdir, src=entry.path, dest=item.dest + [leafdir.dest]
        )



from .request import (
    inspect,
    get_manifest,
    api_version_check,
    AuthToken,
    MEDIATYPE_SCHEMA2_V1,
    MEDIATYPE_SCHEMA2_V1_SIGNED,
    MEDIATYPE_SCHEMA2_V2,
    MEDIATYPE_SCHEMA2_V2_LIST,
)
